/*
 * JNU - Java Native Utility stuff.
 *
 * The +name+ JNU is from Liang, The Java Native Interface, Addison Wesley 1999,
 * where it refers to a family of 'C' functions.
 * The C++ classes herein are +my own+, in both design and implementation.
 *
 * Note that none of these classes stores a JNIEnv pointer,
 * which means that they are all thread-safe.
 * Instead the JNIEnv is always acquired from the JVM on demand.
 *
 * They are all primarily designed to be used on the stack,
 * rather than via new and delete, which is strongly discouraged
 *
 * This work is not to be confused with the article and source code by
 * Gabrilovich and Finkelstein (C/C++ Users Journal, 19(1):10-21, January 2001).
 * Compared with that, this work:
 *
 *	- is somewhat simpler ;-).
 *	- supports a Java-like style of programming, e.g. by providing Java APIs
 *		where possible, rather than providing a C++-like style with
 *		widespread use of constructors and overloaded = operators.
 *	- is considerably more free of bugs ;-)).
 *	- avoids the resource-copying issue altogether by simply prohibiting
 *		the copying of classes which allocate resources behind the scenes.
 *	- relies exclusively on C++ classes and templates,
 *		and makes no use of the preprocessor whatsoever.
 *	- provides considerably more coverage of the JNI API.
 *
 * Typical usage: a method that just calls self.setValue(value):
 *
 * JNIEXPORT void JNICALL Java_com_mycorp_mypackage_myclass_mynative0(JNIEnv *env, jobject self, jint value)
 * {
 *	JNULocalFrame	frame;
 *	JNUObject<jobject> jnuSelf = self;
 *  JNUClass  clazz = jnuSelf.getClass();
 *  JNUVoidMethod method = clazz.getMethod("setValue","(I)V");
 *	method.invoke(jnuSelf, value);
 * }
 *
 * You get the idea. It is very much like Java reflection.
 *
 * Copyright (c) Esmond Pitt, 2008. All rights reserved.
 */

#include <assert.h>
#include <memory.h>
#include <stdarg.h>
#include <exception>

#include "jni.h"

#ifdef _MSC_VER
#pragma warning(push, 1)
//#pragma warning(disable : 4786)
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif


 
// Forward references
namespace java
{
	namespace jnu
	{

		class JNUJVM;			// = JavaVM;
		class JNUEnv;			// = JNIEnv
		class JNULocalFrame;	// JNI Push/PopLocalFrame
	};
	namespace lang
	{

		// JType is the JNI type of the object: jobject, jclass, jarray etc
		template<class JType> class JNUObject;	// = java.lang.Object
		class JNUClass;							// = java.lang.Class
		class JNUThrowable;						// = java.lang.Throwable.
		class JNUString;						// = java.lang.String

		namespace ref
		{
			class JNUGlobalRef;
			class JNULocalRef;
			template<class JType> class JNUWeakGlobalRef;
		}
		namespace reflect
		{
			template<class T>class JNUMemberID;
			/*
			 * JNUFieldID wraps jfieldID but also contains the jclass the fieldID came from.
			 * JNUMethodID wraps jmethodID but also contains the jclass the methodD came from.
			 */
			typedef JNUMemberID<jfieldID>	JNUFieldID;
			typedef JNUMemberID<jmethodID>	JNUMethodID;

			template<class JType> class JNUMember;	// = java.lang.reflect.Member
			template<class JType> class JNUField;	// = java.lang.reflect.Field

			// Template instances of JNUField for all the types.
			typedef JNUField<jboolean>	JNUBooleanField;
			typedef JNUField<jbyte>		JNUByteField;
			typedef JNUField<jchar>		JNUCharField;
			typedef JNUField<jdouble>	JNUDoubleField;
			typedef JNUField<jfloat>	JNUFloatField;
			typedef JNUField<jint>		JNUIntField;
			typedef JNUField<jlong>		JNULongField;
			typedef JNUField<jshort>	JNUShortField;
			typedef JNUField<jobject>	JNUObjectField;

			// The JType is the return type of the method.
			template<class JType> class JNUMethod;		// = java.lang.reflect.Method

			// Template instances of JNUMethod for all the types.
			typedef JNUMethod<jboolean>	JNUBooleanMethod;
			typedef JNUMethod<jbyte>	JNUByteMethod;
			typedef JNUMethod<jchar>	JNUCharMethod;
			typedef JNUMethod<jdouble>	JNUDoubleMethod;
			typedef JNUMethod<jfloat>	JNUFloatMethod;
			typedef JNUMethod<jint>		JNUIntMethod;
			typedef JNUMethod<jlong>	JNULongMethod;
			typedef JNUMethod<jshort>	JNUShortMethod;
			typedef JNUMethod<void>		JNUVoidMethod;
			typedef JNUMethod<jobject>	JNUObjectMethod;

			class JNUConstructor;						// = java.lang.reflect.Constructor
		}; // java.lang.reflect

		// The JType is the type of the array: jbooleanArray etc.
		template<class JType> class JNUArray;			// = jarray
		// The JType is the type of the array: jbooleanArray etc.
		// The NativeType is the underlying JNI type: jboolean etc.
		template<class JArrayType, class NativeType> class JNUPrimitiveArray;
		
		// Template instances of JNUPrimitiveArray for all the primitive types.
		typedef class JNUPrimitiveArray<jbooleanArray,	jboolean>	JNUBooleanArray;
		typedef class JNUPrimitiveArray<jbyteArray,		jbyte>		JNUByteArray;
		typedef class JNUPrimitiveArray<jcharArray,		jchar>		JNUCharArray;
		typedef class JNUPrimitiveArray<jdoubleArray,	jdouble>	JNUDoubleArray;
		typedef class JNUPrimitiveArray<jfloatArray,	jfloat>		JNUFloatArray;
		typedef class JNUPrimitiveArray<jintArray,		jint>		JNUIntArray;
		typedef class JNUPrimitiveArray<jlongArray,		jlong>		JNULongArray;
		typedef class JNUPrimitiveArray<jshortArray,	jshort>		JNUShortArray;

		// The JType is the type of the array: jbooleanArray etc.
		// The NativeType is the underlying JNI type: jboolean etc.
		template<class JArrayType, class NativeType> class JNUPrimitiveArrayCritical;
		
		// Template instances of JNUPrimitiveArrayCritical for all the primitive types.
		typedef class JNUPrimitiveArrayCritical<jbooleanArray,	jboolean>	JNUBooleanArrayCritical;
		typedef class JNUPrimitiveArrayCritical<jbyteArray,		jbyte>		JNUByteArrayCritical;
		typedef class JNUPrimitiveArrayCritical<jcharArray,		jchar>		JNUCharArrayCritical;
		typedef class JNUPrimitiveArrayCritical<jdoubleArray,	jdouble>	JNUDoubleArrayCritical;
		typedef class JNUPrimitiveArrayCritical<jfloatArray,	jfloat>		JNUFloatArrayCritical;
		typedef class JNUPrimitiveArrayCritical<jintArray,		jint>		JNUIntArrayCritical;
		typedef class JNUPrimitiveArrayCritical<jlongArray,		jlong>		JNULongArrayCritical;
		typedef class JNUPrimitiveArrayCritical<jshortArray,	jshort>		JNUShortArrayCritical;
		
		class JNUObjectArray;	// = jobjectArray, java.lang.Object[]
		
	};	// java.lang

	namespace nio
	{
		class JNUBuffer;		// = java.nio.Buffer
		class JNUByteBuffer;	// = java.nio.ByteBuffer
	}; // java.nio
};
using namespace std;
using namespace java::lang;
using namespace java::lang::ref;
using namespace java::lang::reflect;
using namespace java::nio;
using namespace java::jnu;

namespace java { namespace jnu {


/*
 * PendingJavaException is thrown whenever a JNI exception is pending.
 * It means that the calling JNI method should exit ASAP
 * back to the JVM without calling any more JNI routines
 * (which is dangerous when a Java exception is pending);
 * then the JVM will throw the pending JNI exception to the Java caller.
 * This exception contains nothing of interest:
 * hence it is usually caught with 'catch (PendingJavaException &)'
 * followed by 'return' or 'return  0;'.
 */
class PendingJavaException : public exception {};

/*
 * JNUJVM stores a reference to the current JVM and makes
 * available the current JNIEnv.
 */
class JNUJVM
{
public:
	/*
	 * Return the JNUEnv for the current thread.
	 */
	static JNUEnv env();
	/*
	 * Get a list of the current JVMs.
	 * At present there will only be zero or one.
	 */
	static jint	getCreatedJavaVMs(JavaVM *&jvms, jint size)
		throw(PendingJavaException);
	/*
	 * Get the JVM version.
	 */
	static jint	getVersion();

private:
	friend JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *jvm, void *reserved);
	friend JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *jvm, void *reserved);

	static JavaVM	*jvm;
};

/*
 * JNUEnv stores a JNIEnv * pointer and provides some other useful stuff.
 */
class JNUEnv
{
public:
	typedef JNUEnv	This;
	/*
	 * Constructor.
	 */
	JNUEnv(JNIEnv *env);
	
	/*
	 * Destructor.
	 */
	virtual ~JNUEnv();

	/*
	 * Cast myself to the JNIEnv *
	 */
	operator JNIEnv *() const	{ return _env;	}

	JNIEnv *operator->() const	{ return _env;	}

	/*
	 * Maintain a weak global reference to the class of the specified object.
	 * The 'weak' parameter should be a static object: it is maintained by this method.
	 * The method returns a local reference to the object, or 0 if
	 * the weak global reference has been garbage-collected.
	 */
	jclass getWeakGlobalRef(const JNUObject<jobject> &object, jweak &weak);

	/*
	 * Return a weak global reference to the specified class.
	 * The 'weak' parameter should be a static object: it is maintained by this method.
	 */
	jclass getWeakGlobalRef(const char *className, jweak &weak);

	/*
	 * Create and throw an exception into the Java VM using this wrapper.
	 * If the specified exception class can't be found, the VM will throw
	 * a class not found exception.
	 */
	void throwException(const char *excClassName, const char *message);

	/*
	 * Check for a pending Java exception
	 */
	void	checkPendingException() const
		throw(PendingJavaException);

protected:
	JNIEnv	*_env;
};

/*
 * JNULocalFrame wraps Push/PopLocalFrame,
 * ensuring release of all local references acquired in the same scope.
 * Declare one of these, *on the stack* (not via new()),
 * in every JNI method that acquires local references.
 * When the enclosing JNI scope exits, by any means, the destructor of this
 * class will ensure the local references acquires in the scope are released.
 *
 * *If the JNI method returns a jobject or jclass or jobjectarray or jarray*,
 * the reference has to be passed through PopLocalFrame() to associate it with
 * the next outer local frame. In this class, this is accomplished by the pop()
 * method, passing the object reference required: a new value for the enclosing frame
 * is returned.
 * Note that this also calls PopLocalFrame() 'prematurely',
 * invalidating all local references acquired within the scope except for the
 * one passed to pop(), but PopLocalFrame() won't be called again at scope exit.
 * This curious design flows from the design of PopLocalFrame(), sorry.
 */
class JNULocalFrame : public JNUEnv
{
public:
	typedef JNULocalFrame	This;
	
	/* The JNI default frame size is 16. */

	/*
	 * Construct a local frame of the specified capacity.
	 */
	JNULocalFrame(jint capacity = 16)
		throw(PendingJavaException);

	/*
	 * Destroy this local frame, invalidating all the local references
	 * acquired in the same scope.
	 */
	virtual ~JNULocalFrame();

private:
	// private copy ctor & assignment operator, i.e no copying allowed.
	JNULocalFrame(const JNULocalFrame &);
	JNULocalFrame &operator=(const JNULocalFrame &);

public:
	/*
	 * Ensures that at least a given number of local references can be created
	 * in the current thread. Throws an OutOfMemoryError via a PendingJavaException.
	 */
	static void	ensureLocalCapacity(jint capacity)
		throw(PendingJavaException);
	/*
	 * Pop the specified object out of this local frame,
	 * e.g. so it can be returned as a result by the enclosing JNI method.
	 * This has the 'side-effect' of invalidating all other local references
	 * acquired in the scope of this frame, so it is used thus:
	 *
	 *	JNULocalFrame	frame(env);
	 *	// create 'result' and lots of other local refs
	 *	// only want to return 'result', lose all the other local refs.
	 *	return frame.pop(result);
	 *
	 * T must be jobject, jclass, jobjectArray, jintArray, etc.
	 * Otherwise you will get compile errors. Sorry, no generics in C++.
	 */
	template<class T>	T	pop(T object) throw(PendingJavaException)
	{
		// printf("popping a local variable %8.8x\n", object); fflush(stdout);
		_popped = 1;
		T	result = (T)_env->PopLocalFrame(object);
		_env->ExceptionDescribe();
		return result;
	}

private:
	bool	_popped;			// true if pop() has already been called
};

};	// namespace java.jnu

namespace lang	// java.lang
{
/*
 * JNUObject is a base class like java.lang.Object.
 * The JType parameter is usually one of:
 * jobject, jstring,
 * jarray or one of its subtypes,
 * jfieldID, jmethodID, ...
 *
 * To be like Java it provides provides a getClass() method.
 * It does nothing else much except define a virtual destructor
 * and an operator to cast itself to its JType.
 */
template<class JType> class JNUObject
{
public:
	typedef JNUObject<JType>	This;
	/*
	 * Constructor.
	 */
	JNUObject(JType object)
		throw(PendingJavaException)
		: _object(object)
	{
		/*
		* Don't do this. JNUString allows null jstrings.
		if (object == 0)
		{
		JNUJVM::env().throwException("java/lang/NullPointerException", "constructing JNUObject");
		throw PendingJavaException();
		}
		assert(object != 0);
		*/
	}


	/*
	 * Destructor.
	 */
	virtual ~JNUObject() throw() {}

	/*
	 * Return the class of the object.
	 */
	virtual JNUClass	getClass() const;

	/*
	 * Cast myself to the JType.
	 */
	virtual operator JType() const	{ return _object;	}

	/*
	 * Return the JType object.
	 */
	virtual JType	object() const	{ return _object;	}

	/*
	 * Identity-checking operator==.
	 */
	virtual bool	operator==(const JNUObject<JType> &that) const
	{
		return JNUJVM::env()->IsSameObject(*this, that);
	}

protected:
	void setObject(JType object) const
	{
		const_cast<JNUObject*>(this)->_object = object;
	}
	JType	_object;
};

/*
 * synchronized wraps MonitorEnter/MonitorExit,
 * making it look somewhat like a Java synchronized() block.
 * However unlike a Java synchronized() block, the synchronized(object)
 * statement needs to be place +inside+ the block in question, e.g.
 *
 *	{
 *		synchronized(object);
 *		// ...
 *	}
 */
class synchronized
{
public:
	typedef synchronized	This;

	synchronized(const JNUObject<jobject> &object);

	virtual ~synchronized();

private:
	const JNUObject<jobject>	&_object;
};

/*
 * JNUClass wraps FindClass(), GetObjectClass(), and getMethod(),
 * throwing PendingJavaException any time there is a pending JNI exception.
 *
 * Like Java, it provides a JNUClass::forName() method.
 */
class JNUClass : public JNUObject<jclass>
{
public:
	typedef JNUClass	This;
	/*
	 * Construct a JNUClass from a jclass
	 */
	JNUClass(jclass clazz)
		throw(PendingJavaException);

	/*
	 * Construct a JNUClass from a class name
	 */
	static JNUClass	forName(const char *className)
		throw(PendingJavaException);

	/*
	 * Construct a JNUClass from a class name and a weak global reference to it,
	 * which is assumed to be static.
	 */
	static JNUClass	forName(const char *className, jweak &cls)
		throw(PendingJavaException);

	/*
	 * Get the specified constructor.
	 * The default signature argument refers to the default (no-args) constructor.
	 */
	JNUConstructor	getConstructor(const char *signature = "()V") const
		throw(PendingJavaException);

	/*
	 * Return the fieldID of the specified instance field.
	 * You can convert this to a JNUField<type> of the appropriate type
	 * simply by assignment.
	 */
	JNUFieldID	getField(const char *name, const char *signature) const
		throw(PendingJavaException);

	/*
	 * Return the fieldID of the specified static field.
	 * You can convert this to a JNUField<type> of the appropriate type
	 * simply by assignment.
	 */
	JNUFieldID getStaticField(const char *name, const char *signature) const
		throw(PendingJavaException);

	/*
	 * Return the methodID of the specified method.
	 * You can convert this to a JNUMethod<type> of the appropriate type
	 * simply by assignment.
	 */
	JNUMethodID	getMethod(const char *name, const char *signature) const
		throw(PendingJavaException);

	/*
	 * Return the methodID of the specified static method.
	 * You can convert this to a JNUMethod<type> of the appropriate type
	 * simply by assignment.
	 */
	JNUMethodID	getStaticMethod(const char *name, const char *signature) const
		throw(PendingJavaException);

#if NOTYET	// Not working yet for reasons unknown.
	/*
	 * Return the name of the class.
	 */
	virtual JNUString	getName() const;
#endif

	/*
	 * Return the superclass of this class, or null if this is JNUObject.clazz.
	 */
	jclass	getSuperclass() const;

	/*
	 * Return true if this class is assignable from the specified class.
	 */
	jboolean	isAssignableFrom(const JNUClass &clazz) const;

	/*
	 * Return true if the specified object is an instance of this class.
	 */
	jboolean	isInstance(const jobject object) const;

	/*
	 * Construct a new instance of the class.
	 * This calls the no-arg constructor.
	 * The current JNUClass must not refer to an array class.
	 */
	jobject	newInstance() const
		throw(PendingJavaException);
};

template<class JType> JNUClass	JNUObject<JType>::getClass() const
{
	JNUEnv	env = JNUJVM::env();
	JNUClass	clazz = JNUClass(env->GetObjectClass(object()));
	// Exception can only happen if object is NULL, as all objects have classes.
	env.checkPendingException();
	return clazz;
}

namespace ref
{

// Base reference class
class JNURef
{
protected:
	JNURef(jobject ref);
	virtual ~JNURef();
public:
	virtual operator jobject() const	{ return _ref;	}

protected:
	jobject	_ref;
};

/*
 * JNUGlobalRef holds a global reference to a jobject or jclass.
 * The global reference is released when the object is destroyed.
 */
class JNUGlobalRef : public JNURef
{
public:
	JNUGlobalRef(jobject ref);
	JNUGlobalRef(const JNUGlobalRef &that);
	JNUGlobalRef operator=(const JNUGlobalRef &that);

	~JNUGlobalRef();

	/*
	 * Convert the global reference to a local reference attached
	 * to the current thread.
	 * The global reference can't be used directly,
	 * it just exists to keep the object
	 * out of the garbage-collector's dastardly clutches.
	 * The local reference allocated here is released when the returned
	 * JNULocalRef is destroyed.
	 */
	JNULocalRef	getLocalRef() const;
};

/*
 * JNULocalRef holds a local reference to a jobject or jclass.
 * The local reference is released when the object is destroyed.
 */
class JNULocalRef : public JNURef
{
public:
	// Constructor
	JNULocalRef(jobject ref);
	// Destructor
	virtual ~JNULocalRef();
protected:
	// You can't copy these around between threads.
	JNULocalRef(const JNULocalRef &that);
	JNULocalRef operator=(const JNULocalRef &that);

	friend class JNUGlobalRef;
};

/*
 * JNUWeakGlobalRef allows the caller to hold a weak reference to a jclass or jobject.
 * It is used thus:
 *
 * static jweak	cls;
 * JNUWeakGlobalRef<JNUClass>	jcwf(env, "classname", cls);
 *
 * Thereafter jcwf can be used as a jclass object, or as a JNUClass object.
 *
 */
template<class JType> class JNUWeakGlobalRef : public JNULocalRef
{
private:
	typedef JNUWeakGlobalRef<JType>	This;
	friend class java::lang::JNUObject<jobject>;
	friend class java::lang::JNUClass;

	// Constructors
	JNUWeakGlobalRef(jweak &weak);
	JNUWeakGlobalRef(const char *className, jweak &cls);

	// Copy constructor
	JNUWeakGlobalRef(const JNUWeakGlobalRef<JType> &that);

	virtual ~JNUWeakGlobalRef();

	// Assignment operator.
	JNUWeakGlobalRef	operator=(const JNUWeakGlobalRef<JType> &that);

public:
	// Return this object as a JType, taking care of renewing
	// any weak global references.
	virtual operator JType()	const
		throw(PendingJavaException);

private:
	const char	*_className;
	jweak	&_weak;
};

// Assignment operator.
template<class JType>
JNUWeakGlobalRef<JType>
JNUWeakGlobalRef<JType>::operator=(const JNUWeakGlobalRef<JType> &that)
{
	return JNUWeakGlobalRef<JType>(that);
}

// typecast operator
template<class JType>
JNUWeakGlobalRef<JType>::operator JType() const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	if (_weak == 0 || env->IsSameObject(_weak, 0))
		return 0;
	return (JType)_ref;
}

// Specialization of typecast operator for jclass
template<> JNUWeakGlobalRef<jclass>::operator jclass() const
	throw(PendingJavaException);

};	// namespace java.lang.ref

/*
 * JNUString wraps NewString()/NewStringUTF(), Get/ReleaseStringChars(),
 * Get/ReleaseStringUTFChars(), and GetStringRegion()/GetStringUTFRegion(),
 * ensuring release of any acquired data.
 * Declare one of these, *on the stack* (not via new()),
 * for each jstring you want to get a 'jchar *' or 'char *' from.
 * When the enclosing JNI scope exits, by any means, the destructor of this
 * class will ensure the aqcuired data is released.
 */
class JNUString : public JNUObject<jstring>
{
public:
	typedef JNUString	This;
	typedef JNUObject<jstring>	super;
	/*
	 * Construct a JNUString object from the specified jstring.
	 * 'string' may be null, in which case the class functions like (const char*)0.
	 */
	JNUString(jstring string);

	/*
	 * Construct a JNUString object from the specified unicode jchars.
	 */
	JNUString(const jchar *unicode, jsize length);

	/*
	 * Construct a JNUString object from the specified JNIEnv and UTF chars.
	 * 'string' may be null, in which case the class functions like (const char*)0.
	 */
	JNUString(const char *utf);

	/*
	 * Destroy a JNUString object, releasing the char * data.
	 */
	virtual ~JNUString() throw();

private:
	// private copy ctor & assignment operator, i.e no copying allowed.
	JNUString(const JNUString&);
	JNUString&operator=(const JNUString&);

public:
	/*
	 * Return the length of the string in Unicode chars (jhcars).
	 * This is the same as java.lang.String.length().
	 * Returns zero if the jstring was null.
	 */
	jsize	getLength() const;

	/*
	 * Return the chars.
	 */
	const jchar *getChars();
	/*
	 * Cast myself to the jchars.
	 */
	operator const jchar *()	{ return getChars();	}

	/*
	 * Fetch a region of the string into the specified jchar buffer.
	 */
	void getRegion(jsize start, jsize len, jchar *buf);

	/*
	 * Return the UTF.
	 */
	const char *getUTFChars();

	/*
	 * Return the length of the UTF chars that would be required to hold this string.
	 * Note that this can be longer than getLength().
	 * Returns zero if the jstring was null.
	 */
	jsize	getLengthUTF() const;

	/*
	 * Cast myself to the UTF chars.
	 */
	operator const char *()	{ return getUTFChars();	}

	/*
	 * Fetch a region of the string into the specified UTF buffer.
	 */
	void getRegion(jsize start, jsize len, char *buf);

private:
	// Pointers to data acquired from the jstring.
	// Note: these never point to data supplied by the application
	// via a constructor.
	const jchar		*_unicode;	// The acquired jchar data, or zero.
	const char		*utf;		// The acquired UTF data, or zero.
};

/*
 * JNUThrowable corresponds to java.lang.Throwable.
 */
class JNUThrowable : public JNUObject<jthrowable>, public PendingJavaException
{
public:
	JNUThrowable(jthrowable throwable);
	/*
	 * Constructing one of these via this constructor throws the named Java exception.
	 * You can then throw it as a C++ exception.
	 */
	JNUThrowable(const char *className, const char *message = 0);

	/*
	 * Throw this exception. Can't call it throw(), sorry.
	 */
	void throwIt();

	/*
	 * Print the stack trace of the current Java exception, if any.
	 * NB that's not necessarily this one! Sorry!
	 */
	static void	printStackTrace();

	/*
	 * Return true if a Java exception is pending.
	 */
	static jboolean	isExceptionPending();

	/*
	 * Clear any current Java exception.
	 */
	static void	clear();

	/*
	 * Return the current Java exception, or null.
	 */
	static jthrowable	currentException();
};

// Some handy exceptions.
class JNUException : public JNUThrowable
{
public:
	JNUException(const char *className, const char *message = 0)
		: JNUThrowable(className, message)
	{
	}
	JNUException(const char *message = 0)
		: JNUThrowable("java/lang/Exception", message)
	{
	}
};

class JNURuntimeException : public JNUException
{
public:
	JNURuntimeException(const char *className, const char *message = 0)
		: JNUException(className, message)
	{
	}
	JNURuntimeException(const char *message = 0)
		: JNUException("java/lang/RuntimeException", message)
	{
	}
};

class JNUIllegalArgumentException : public JNURuntimeException
{
public:
	JNUIllegalArgumentException(const char *message = 0)
		: JNURuntimeException("java/lang/IllegalArgumentException", message)
	{
	}
};

class JNUNullPointerException : public JNURuntimeException
{
public:
	JNUNullPointerException(const char *message = 0)
		: JNURuntimeException("java/lang/NullPointerException", message)
	{
	}
};

class JNUUnsupportedOperationException : public JNURuntimeException
{
public:
	JNUUnsupportedOperationException(const char *message = 0)
		: JNURuntimeException("java/lang/UnsupportedOperationException", message)
	{
	}
};

class JNUIOException : public JNUException
{
public:
	JNUIOException(const char *message = 0)
		: JNUException("java/io/IOException", message)
	{
	}
};

namespace reflect
{

template<class T>class JNUMemberID
{
protected:
	friend class java::lang::JNUClass;
	
	JNUMemberID(const char *name, T id, jclass clazz, bool isStatic = JNI_FALSE)
		: _name(name),
		_id(id),
		_clazz(clazz),
		_isStatic(isStatic)
	{
	}

public:
	// Copy ctor
	JNUMemberID(const JNUMemberID<T> &that)
		: _name(that._name),
		_id(that._id),
		_clazz(that._clazz),
		_isStatic(that._isStatic)
	{
	}

public:
	const char			*getName() const	{ return _name;	}
	virtual T			id() const			{ return _id; }
	virtual operator	T() const			{ return _id; }
	virtual operator	jclass() const		{ return _clazz; }
	virtual bool		isStatic() const	{ return _isStatic;	}

private:
	const char	*_name;
	const T	_id;
	jclass	_clazz;
	bool	_isStatic;
};

/*
 * JNUMember corresponds to the java.lang.reflect.Member interface.
 */
template<class IDType> class JNUMember : protected JNUMemberID<IDType>
{
public:
	typedef JNUMember<IDType>	This;
protected:
	JNUMember(const JNUMemberID<IDType> &id)
		: JNUMemberID<IDType>(id)
	{
	}
public:
	/*
	 * @see java.lang.reflect.Method.getDeclaringClass().
	 */
	virtual JNUClass	getDeclaringClass() const
	{
		return (jclass)*this;
	}
};

/*
 * JNUAccessibleObject corresponds to java.lang.reflect.AccessibleObject.
 * It extends JNUObject, and ties together
 * JNUField, JNUMethod, and JNUConstructor.
 */
template<class IDType> class JNUAccessibleObject : public JNUObject<jobject>
{
public:
	typedef JNUAccessibleObject<IDType>	This;
protected:
	JNUAccessibleObject(const JNUMemberID<IDType> &id)
		: JNUObject<jobject>(0),
		_id(id)
	{
	}
public:
	/*
	 * Override JNUObject::getClass to return Field.class, Method.class, Constructor.class
	 * This calls the getClass() method below and the same remarks apply.
	 */
	virtual JNUClass	getClass() const
	{
		return JNUObject<jobject>(*this).getClass();
	}

	/*
	 * Turn this JNUAccessibleObject into a java.lang.reflect.Field or Method or Constructor as appropriate.
	 * This operation requires a JNI call, and allocates a local reference
	 * to the object, so it has its costs.
	 * Calling it isn't necessary for any purpose I can
	 * think of but it's provided for completeness, so that JNUMember
	 * really does extend JNUObject correctly.
	 */
	virtual operator	jobject() const
	{
		if (object() == 0)
		{
			setObject((JNUJVM::env()->*ToReflected)(_id, _id, _id.isStatic()));
		}
		return object();
	}
private:
	const JNUMemberID<IDType> &_id;
	static jobject (JNIEnv::*const ToReflected)(jclass, IDType, jboolean);
};


/*
 * JNUField wraps the family of Get<Type>Field()/Set<Type>Field() methods.
 * Type is the native JNI type: jboolean, jbyte, ... jobject.
 */
template<class Type> class JNUField
	: public JNUAccessibleObject<jfieldID>,
	public JNUMember<jfieldID>
{
public:
	typedef JNUField<Type>	This;

	/*
	 * Constructor.
	 */
	JNUField(const JNUFieldID &fieldID);

	/*
	 * Get the value of an instance field.
	 * 'object' is the instance containing the field.
	 */	 
	Type	get(const JNUObject<jobject> &object) const
		throw(PendingJavaException);

	/*
	 * Get the value of a static field.
	 * 'clazz' is the class containing the static field.
	 */	 
	Type	get(const JNUClass &clazz) const
		throw(PendingJavaException);

	/*
	 * Set the value of an instance field.
	 * 'object' is the instance containing the field.
	 */	 
	void	set(jobject object, Type value)
		throw(PendingJavaException);

	/*
	 * Set the value of an instance field.
	 * 'object' is the instance containing the field.
	 * This overload is required because the compiler won't
	 * automatically construct a non-const JNUObject from a jobject.
	 */	 
	void	set(JNUObject<jobject> &object, Type value)
		throw(PendingJavaException);

	/*
	 * Set the value of a static field.
	 * 'clazz' is the Class containing the static field.
	 */	 
	void	set(jclass clazz, Type value)
		throw(PendingJavaException);

	/*
	 * Set the value of a static field.
	 * 'clazz' is the Class containing the static field.
	 * This overload is required because the compiler won't
	 * automatically construct a non-const JNUClass from a jclass.
	 */	 
	void	set(JNUClass &clazz, Type value)
		throw(PendingJavaException);

private:
	static Type (JNIEnv::*const GetField)(jobject, jfieldID);
	static Type (JNIEnv::*const GetStaticField)(jclass, jfieldID);
	static void (JNIEnv::*const SetField)(jobject, jfieldID, Type);
	static void (JNIEnv::*const SetStaticField)(jclass, jfieldID, Type);
};

template<class Type>
JNUField<Type>::JNUField(const JNUFieldID &fieldID)
	: JNUAccessibleObject<jfieldID>(fieldID),
	JNUMember<jfieldID>(fieldID)
{
}

template<class Type>
Type
JNUField<Type>::get(const JNUObject<jobject> &object) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	Type	value = (env->*GetField)(object, id());
	env.checkPendingException();
	return value;
}

template<class Type>
Type
JNUField<Type>::get(const JNUClass &clazz) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	Type	value = (env->*GetStaticField)(clazz, id());
	env.checkPendingException();
	return value;
}

template<class Type>
void
JNUField<Type>::set(jobject object, Type value)
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	(env->*SetField)(object, id(), value);
	env.checkPendingException();
}

template<class Type>
void
JNUField<Type>::set(JNUObject<jobject> &object, Type value)
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	(env->*SetField)(object, id(), value);
	env.checkPendingException();
}

template<class Type>
void
JNUField<Type>::set(jclass clazz, Type value)
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	(env->*SetStaticField)(clazz, id(), value);
	env.checkPendingException();
}

template<class Type>
void
JNUField<Type>::set(JNUClass &clazz, Type value)
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	(env->*SetStaticField)(clazz, id(), value);
	env.checkPendingException();
}

/*
 * JNUMethod wraps the family of Call<Type>Method[AV]methods.
 * Type is the return type: jboolean, jbyte, ... void, jobject.
 */
template<class Type> class JNUMethod : public JNUAccessibleObject<jmethodID>, public JNUMember<jmethodID>
{
public:
	typedef JNUMethod<Type>	This;

	/*
	 * Constructor.
	 */
	JNUMethod(const JNUMethodID &methodID)
		: JNUAccessibleObject<jmethodID>(methodID),
		JNUMember<jmethodID>(methodID)
	{
	}

	/*
	 * Call an instance method with varargs.
	 * NOTE: this method can't be defined with const JNUObject &
	 * in place of jobject, because the MSVC varargs va_start() macro
	 * doesn't work correctly with reference arguments.
	 */	 
	Type	invoke(jobject object, ...) const
		throw(PendingJavaException);

	/*
	 * Call an instance method with an arglist of jvalue.
	 */	 
	Type	invoke(const JNUObject<jobject> &object, const jvalue *args) const
		throw(PendingJavaException);

	/*
	 * Call an instance method with a va_list.
	 */	 
	Type	invoke(const JNUObject<jobject> &object, va_list args) const
		throw(PendingJavaException);

	/*
	 * Call a static method with varargs
	 * NOTE: this method can't be defined with const JNUClass &
	 * in place of jclass, because the MSVC varargs va_start() macro
	 * doesn't work correctly with reference arguments.
	 */	 
	Type	invoke(jclass clazz, ...) const
		throw(PendingJavaException);

	/*
	 * Call a static method with an arglist of jvalues.
	 */	 
	Type	invoke(const JNUClass &clazz, const jvalue *args) const
		throw(PendingJavaException);

	/*
	 * Call a static method with a va_list
	 */	 
	Type	invoke(const JNUClass &clazz, va_list args) const
		throw(PendingJavaException);

	/*
	 * Call an instance method nonvirtually via the specified object and class with varargs.
	 * NOTE: this method can't be defined with const JNUClass &
	 * in place of jclass, because the MSVC varargs va_start() macro
	 * doesn't work correctly with reference arguments.
	 * 'clazz' is the class in the inheritance hierarchy whose method it is desired to call.
	 */
	Type	invokeNonVirtual(const JNUObject<jobject> &object, jclass clazz, ...) const
		throw(PendingJavaException);

	/*
	 * Call an instance method nonvirtually via the specified object and class with an arglist of jvalues.
	 * 'clazz' is the class in the inheritance hierarchy whose method it is desired to call.
	 */	 
	Type	invokeNonVirtual(const JNUObject<jobject> &object, const JNUClass &clazz, const jvalue *args) const
		throw(PendingJavaException);

	/*
	 * Call an instance method nonvirtually via the specified object and class with a va_list.
	 * 'clazz' is the class in the inheritance hierarchy whose method it is desired to call.
	 */	 
	Type	invokeNonVirtual(const JNUObject<jobject> &object, const JNUClass &clazz, va_list args) const
		throw(PendingJavaException);

private:
//	static Type (JNIEnv::*const CallMethod)(jobject, jmethodID, ...);
	static Type (JNIEnv::*const CallMethodA)(jobject, jmethodID, const jvalue *);
	static Type (JNIEnv::*const CallMethodV)(jobject, jmethodID, va_list);
//	static Type (JNIEnv::*const CallStaticMethod)(jobject, jmethodID, ...);
	static Type (JNIEnv::*const CallStaticMethodA)(jclass, jmethodID, const jvalue *);
	static Type (JNIEnv::*const CallStaticMethodV)(jclass, jmethodID, va_list);
//	static Type (JNIEnv::*const CallNonvirtualMethod)(jobject, jmethodID, ...);
	static Type (JNIEnv::*const CallNonvirtualMethodA)(jobject, jclass, jmethodID, const jvalue *);
	static Type (JNIEnv::*const CallNonvirtualMethodV)(jobject, jclass, jmethodID, va_list);

};

template<class Type>	Type	JNUMethod<Type>::invoke(jobject object, ...) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	va_list	args;
	va_start(args, object);
	Type	result = (env->*CallMethodV)(object, id(), args);
	va_end(args);
	env.checkPendingException();
	return result;
}

template<class Type>	Type	JNUMethod<Type>::invoke(const JNUObject<jobject> &object, const jvalue *args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	Type	result = (env->*CallMethodA)(object, id(), args);
	env.checkPendingException();
	return result;
}

template<class Type>	Type	JNUMethod<Type>::invoke(const JNUObject<jobject> &object, va_list args) const
		throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	Type	result = (env->*CallMethodV)(object, id(), args);
	env.checkPendingException();
	return result;
}

template<class Type>	Type	JNUMethod<Type>::invoke(jclass clazz, ...) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	va_list	args;
	va_start(args, clazz);
	Type	result = (env->*CallStaticMethodV)(clazz, id(), args);
	va_end(args);
	env.checkPendingException();
	return result;
}

template<class Type>	Type	JNUMethod<Type>::invoke(const JNUClass &clazz, const jvalue *args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	Type	result = (env->*CallStaticMethodA)(clazz, id(), args);
	env.checkPendingException();
	return result;
}

template<class Type>	Type	JNUMethod<Type>::invoke(const JNUClass &clazz, va_list args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	Type	result = (env->*CallStaticMethodV)(clazz, id(), args);
	env.checkPendingException();
	return result;
}

template<class Type>	Type	JNUMethod<Type>::invokeNonVirtual(const JNUObject<jobject> &object, jclass clazz, ...) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	va_list	args;
	va_start(args, clazz);
	Type	result = (env->*CallNonvirtualMethodV)(object, clazz, id(), args);
	va_end(args);
	env.checkPendingException();
	return result;
}

template<class Type>	Type	JNUMethod<Type>::invokeNonVirtual(const JNUObject<jobject> &object, const JNUClass &clazz, const jvalue *args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	Type	result = (env->*CallNonvirtualMethodA)(object, clazz, id(), args);
	env.checkPendingException();
	return result;
}

template<class Type>	Type	JNUMethod<Type>::invokeNonVirtual(const JNUObject<jobject> &object, const JNUClass &clazz, va_list args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	Type	result = (env->*CallNonvirtualMethodV)(object, clazz, id(), args);
	env.checkPendingException();
	return result;
}

// Specialize the invoke() methods for 'void', there being nothing to return.
template<> inline void JNUVoidMethod::invoke(jobject object, ...) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	va_list	args;
	va_start(args, object);
	env->CallVoidMethodV(object, id(), args);
	va_end(args);
	env.checkPendingException();
}

template<> inline void JNUVoidMethod::invoke(const JNUObject<jobject> &object, const jvalue *args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	(env->CallVoidMethodA)(object, id(), args);
	env.checkPendingException();
}

template<> inline void JNUVoidMethod::invoke(const JNUObject<jobject> &object, va_list args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	env->CallVoidMethodV(object, id(), args);
	env.checkPendingException();
}

template<> inline void JNUVoidMethod::invoke(jclass clazz, ...) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	va_list	args;
	va_start(args, clazz);
	env->CallStaticVoidMethodV(clazz, id(), args);
	va_end(args);
	env.checkPendingException();
}

template<> inline void JNUVoidMethod::invoke(const JNUClass &clazz, const jvalue *args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	env->CallStaticVoidMethodA(clazz, id(), args);
	env.checkPendingException();
}

template<> inline void JNUVoidMethod::invoke(const JNUClass &clazz, va_list args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	env->CallStaticVoidMethodV(clazz, id(), args);
	env.checkPendingException();
}

template<> inline void JNUVoidMethod::invokeNonVirtual(const JNUObject<jobject> &object, jclass clazz, ...) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	va_list	args;
	va_start(args, clazz);
	invoke(object, clazz, args);
}

template<> inline void JNUVoidMethod::invokeNonVirtual(const JNUObject<jobject> &object, const JNUClass &clazz, const jvalue *args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	env->CallNonvirtualVoidMethodA(object, clazz, id(), args);
	env.checkPendingException();
}

template<> inline void JNUVoidMethod::invokeNonVirtual(const JNUObject<jobject> &object, const JNUClass &clazz, va_list args) const
	throw(PendingJavaException)
{
	JNUEnv	env = JNUJVM::env();
	env->CallNonvirtualVoidMethodV(object, clazz, id(), args);
	env.checkPendingException();
}

/*
 * JNUConstructor is like java.lang.reflect.Constructor
 */
class JNUConstructor : public JNUAccessibleObject<jmethodID>, public JNUMember<jmethodID>
{
public:
	typedef JNUConstructor	This;
private:
	friend class java::lang::JNUClass;

	JNUConstructor(const JNUMethodID &methodID);

public:
	/*
	 * Create a new object instance with varargs.
	 * NOTE: this method can't be defined with const JNUClass &
	 * in place of jclass, because the MSVC varargs va_start() macro
	 * doesn't work correctly with reference arguments.
	 */
	jobject	newInstance(jclass cls, ...)
		throw(PendingJavaException);

	/*
	 * Create a new object instance with an arg list.
	 */
	jobject	newInstance(const JNUClass &cls, const jvalue *args)
		throw(PendingJavaException);

	/*
	 * Create a new object instance with a va_list.
	 */
	jobject	newInstance(const JNUClass &cls, va_list args)
		throw(PendingJavaException);
};

};	// namespace java.lang.reflect

/*
 * JNUArray is the base array type, corresponding to jarray.
 * It provides the getLength() and getJavaArray() methods.
 * The JArrayType template argument will be one of the derivations of jarray:
 * jintArray, ..., jobjectArray.
 */
template<class JArrayType> class JNUArray : public JNUObject<JArrayType>
{
public:
	typedef JNUArray<JArrayType>	This;
protected:

	/*
	 * Constructor.
	 * If the length is unknown, omit it,and it will be
	 * evaluated lazily on demand.
	 */
	JNUArray(JArrayType array, jsize length = -1);

public:
	/*
	 * Return the Java primitive array object
	 */
	// NB 'this->' in the below is redundant but required by gcc/g++; seems to be a bug.
	JArrayType	getJavaArray()					{ return (JArrayType)this->_object;	}
	const	JArrayType	getJavaArray() const	{ return (JArrayType)this->_object;	}

	/*
	 * The length of the array.
	 * Like Java's array.length member,
	 * this is provided as a public const data member.
	 */
	const jsize	length;
 };

template<class JArrayType> JNUArray<JArrayType>::JNUArray(JArrayType array, jsize length)
	: JNUObject<JArrayType>(array),
	length(length == -1 ? JNUJVM::env()->GetArrayLength(array) : length)
{
	JNUJVM::env().checkPendingException();
}

/*
 * JNUPrimitiveArray provides access to the data and length of primitive arrays.
 * It is a template class with several arguments:
 * JArrayType:	the Java array type, e.g. jintArray.
 * Type:		the JNI type, e.g. jint.
 * NewArray:	the address of the JNI New<Type>Array method.
 * GetArrayElements:		the address of the JNI Get<Type>ArrayElememnts method.
 * ReleaseArrayElements:	the address of the JNI Release<Type>ArrayElements method.
 *
 * Wouldn't it be nice if C++ had Generics so all this could be self-sanity-checked?
 *
 * A set of typedef classes is provided below so you don't have to know all this stuff.
 */
template<class JArrayType, class NativeType>
	class JNUPrimitiveArray : public JNUArray<JArrayType>
{
public:
	typedef JNUPrimitiveArray<JArrayType, NativeType>	This;
	/*
	 * Construct a new JNUPrimitiveArray object from an existing primitive array.
	 * The 'mode' parameter is either zero (default) or JNI_ABORT.
	 * This controls what happens to the native array when the object is destroyed.
	 * A mode of zero copies the data back to the Java array;
	 * a mode of JNI_ABORT prevents the native data from being copied back.
	 * The JNI_COMMIT value is handled separately, via the commit() method.
	 */
	JNUPrimitiveArray(JArrayType array, int mode = 0)
		throw(PendingJavaException);

	/*
	 * Construct a new JNUPrimitiveArray object with a new primitive array of the specified length.
	 * The 'mode' parameter is either zero (default) or JNI_ABORT.
	 * This controls what happens to the native array when the object is destroyed.
	 * A mode of zero copies the data back to the Java array;
	 * a mode of JNI_ABORT prevents the native data from being copied back.
	 * The JNI_COMMIT value is handled separately, via the commit() method.
	 */
	JNUPrimitiveArray(jsize length, int mode = 0)
		throw(PendingJavaException);

	/*
	 * Destructor.
	 * If abort() hasn't been called, the data is copied back into the Java array.
	 * The Java array and the <Type>Buffer, if any, are not released by this operation.
	 */
	virtual ~JNUPrimitiveArray() throw();

private:
	// private copy ctor & assignment operator, i.e no copying allowed.
	JNUPrimitiveArray(const JNUPrimitiveArray& that);
	JNUPrimitiveArray &operator=(const JNUPrimitiveArray& that);

public:
	/*
	 * Return the native data array
	 */
	virtual NativeType	*getElements();

	virtual const	NativeType	*getElements() const
	{
		return const_cast<This*>(this)->getElements();
	}

	/*
	 * Return true if getElements() returned a copy, otherwise it returned the original data.
	 *
	 * If isCopy() returns true:
	 *		calling commit() or abort() has no effect.
	 * If isCopy() returns false:
	 *		if you want your changes to JNI native objects to take effect
	 *		in the Java primitive array, you must call commit();
	 *		conversely, if you +don't+, you must call abort().
	 */
	virtual jboolean	isCopy() const					{ return _isCopy;	}

	/*
	 * Return the current copy-back mode: 0 (default) or JNI_ABORT.
	 */
	virtual jint	getMode() const	{ return _mode;	}

	/*
	 * Commit any changes made so far to the data, i.e. copy
	 * the native data back immediately into the Java array.
	 * You can call this method more than once.
	 * Calling this method doesn't affect what happens at destruction time,
	 * which depends only on whether abort() has or hasn't been called.
	 */
	virtual void	commit() throw(PendingJavaException);

	/*
	 * Abort any changes made so far to the data.
	 * This tells this class not to update the underlying Java array when destroyed.
	 * The default if abort() hasn't been called is to update the Java array.
	 * NB this method does not restore the data to its original contents.
	 * Calling this method after commit() doesn't rollback the commit().
	 * Calling this method with 'false' returns the mode to the default.
	 */
	virtual void	abort(bool abort = true)
	{
		_mode = abort ? JNI_ABORT : 0;
	}

	/*
	 * Get a region of the array.
	 * target: target buffer.
	 * offset: offset +index+ in the array of the first item to copy.
	 * length: Number of items to copy.
	 * Note that 'offset' isn't a byte offset, it's an index,
	 * whose value as a byte offset depends on the NativeType.
	 */
	virtual void	getRegion(NativeType *target, int offset, int length) const
		throw(PendingJavaException);

	// Get an individual element via operator overload.
	virtual const NativeType	&operator[](unsigned n) const
		throw(PendingJavaException)
	{
		return getElements()[n];
	}

	// Get/set an individual element via operator overload.
	virtual NativeType	&operator[](unsigned n)
	{
		return getElements()[n];
	}

	/*
	 * Set a region of the array.
	 * source: source buffer.
	 * offset: offset +index+ in the array of the first item to copy.
	 * length: Number of items to copy.
	 * Note that 'offset' isn't a byte offset, it's an index,
	 * whose value as a byte offset depends on the NativeType.
	 */
	virtual void	setRegion(const NativeType *source, int offset, int length)
		throw(PendingJavaException);

protected:

	NativeType	*_elements;
	jboolean	_isCopy;
	jint	_mode;

	static JArrayType (JNIEnv::*const NewArray)(jsize);
	static NativeType *(JNIEnv::*const GetArrayElements)(JArrayType, jboolean *);
	static void (JNIEnv::*const ReleaseArrayElements)(JArrayType, NativeType *, jint);
	static void (JNIEnv::*const GetArrayRegion)(JArrayType, jsize, jsize, NativeType *);
	static void (JNIEnv::*const SetArrayRegion)(JArrayType, jsize, jsize, const NativeType *);
};

template<class JArrayType, class NativeType> JNUPrimitiveArray<JArrayType,NativeType>::JNUPrimitiveArray(JArrayType array, int mode)
	throw(PendingJavaException)
	: JNUArray<JArrayType>(array),
	_elements(0),
	_mode(mode)
{
//		JNUJVM::env().checkPendingException();
	if (mode != 0 && mode == JNI_ABORT)
		throw JNUThrowable("java/lang/IllegalArgumentException", "JNUPrimitiveArray.<init>: mode");
}

template<class JArrayType, class NativeType> JNUPrimitiveArray<JArrayType,NativeType>::JNUPrimitiveArray(jsize length, int mode)
	throw(PendingJavaException)
	: JNUArray<JArrayType>((JNUJVM::env()->*NewArray)(length), length),
	_elements(0),
	_mode(mode)
{
	JNUJVM::env().checkPendingException();
	if (mode != 0 && mode == JNI_ABORT)
		throw JNUThrowable("java/lang/IllegalArgumentException", "JNUPrimitiveArray.<init>: mode");
}

template<class JArrayType, class NativeType> JNUPrimitiveArray<JArrayType,NativeType>::~JNUPrimitiveArray() throw()
{
	if (_elements)
	{
		JNUEnv	env = JNUJVM::env();
		// Release the copy of array(). JNI_ABORT => there are no changes to it.
		//			fprintf(stderr, "~JNUPrimitiveArray: releasing array elements %8.8x\n", _data);
		// NB 'this->' below is redundant but required by gcc/g++: seems to be a bug.
		(env->*ReleaseArrayElements)((JArrayType)this->_object, _elements, _mode);
		if (env->ExceptionCheck())
		{
			env->ExceptionDescribe();
			// Don't throw exceptions from destructors!
			// throw PendingJavaException();
		}
	}
}

template<class JArrayType, class NativeType> NativeType	*JNUPrimitiveArray<JArrayType,NativeType>::getElements()
{
	if (!_elements)
	{
		_elements = (JNUJVM::env()->*GetArrayElements)((JArrayType)this->_object, &_isCopy);
	}
	return _elements;
}

template<class JArrayType, class NativeType> void	JNUPrimitiveArray<JArrayType,NativeType>::commit() throw(PendingJavaException)
{
	// NB calling this method with JNI_COMMIT does *not* release the array elements,
	// it just copies them back into the Java space.
	JNUEnv	env = JNUJVM::env();
	(env->*ReleaseArrayElements)((JArrayType)this->_object, _elements, JNI_COMMIT);
	env.checkPendingException();
}

template<class JArrayType, class NativeType> void	JNUPrimitiveArray<JArrayType,NativeType>::getRegion(NativeType *target, int offset, int length) const
	throw(PendingJavaException)
{
	if (
		offset < 0
		|| offset >= this->length
		|| length < 0
		|| offset+length > this->length
		)
		throw JNUThrowable("java/lang/IllegalArgumentException", "getRegion: offset or length out of bounds");
	if (_elements)
	{
		for (int i = 0; i < length; i++)
			target[i] = _elements[offset+i];
	}
	else
	{
		(JNUJVM::env()->*GetArrayRegion)(*this, offset, length, target);
	}
}

template<class JArrayType, class NativeType> void	JNUPrimitiveArray<JArrayType,NativeType>::setRegion(const NativeType *source, int offset, int length)
	throw(PendingJavaException)
{
	if (
		offset < 0
		|| offset >= this->length
		|| length < 0
		|| offset+length > this->length
		)
		throw JNUThrowable("java/lang/IllegalArgumentException", "setRegion: offset or length out of bounds");
	if (_elements)
	{
		for (int i = 0; i < length; i++)
			_elements[offset+i] = source[i];
	}
	else
	{
		(JNUJVM::env()->*SetArrayRegion)(*this, offset, length, source);
	}
}

#if NOTYET
/*
 * JNUPrimitiveArrayCritical wraps Get/ReleasePrimitiveArrayCritical,
 * whose documentation should be studied carefully.
 * It can only be used to wrap an existing array object.
 * No other JNI calls or blocking operations may occur
 * during the lifetime of this object.
 */
template<class JArrayType, class NativeType>
	class JNUPrimitiveArrayCritical : public JNUPrimitiveArray<JArrayType, NativeType>
{
public:
	typedef JNUPrimitiveArrayCritical<JArrayType, NativeType>	This;
	/*
	 * Construct a new JNUPrimitiveArrayCritical object from an existing primitive array.
	 * The 'mode' parameter is either zero (default) or JNI_ABORT.
	 * This controls what happens to the native array when the object is destroyed.
	 * A mode of zero copies the data back to the Java array;
	 * a mode of JNI_ABORT prevents the native data from being copied back.
	 * The JNI_COMMIT value is handled separately, via the commit() method.
	 */
	JNUPrimitiveArrayCritical(JArrayType array, int mode = 0)
		throw(PendingJavaException);

	~JNUPrimitiveArrayCritical();

	/*
	 * The commit() operation IS NOT SUPPORTED for critical primitive arrays.
	 * This implementaton throws JNUUnsupportedOperationException.
	 */
	virtual void	commit()
		throw(PendingJavaException);
};
#endif

/*
 * JNUObjectArray is for object arrays rather than primitive arrays.
 * Object arrays are similar but different ;-):
 * they don't have underlying JNI native arrays;
 * they don't have get/setRegion() operations;
 * instead, get() and set() methods are provided, operating on individual array elements.
 */
class JNUObjectArray : public JNUArray<jobjectArray>
{
public:
	typedef JNUObjectArray	This;
	/*
	 * Construct a new JNUObjectArray object from an existing object array.
	 */
	JNUObjectArray(jobjectArray array)
		throw(PendingJavaException);

	/*
	 * Construct a new JNUObjectArray object with a new object array of the specified class and length,
	 * populated with the specified initial value (default NULL).
	 */
	JNUObjectArray(jsize length, const JNUClass &cls, jobject initial = 0)
		throw(PendingJavaException);

	/*
	 * Destructor.
	 */
	virtual ~JNUObjectArray() throw();

	/*
	 * Get the nth element of the array
	 */
	jobject	get(jsize n) const
		throw(PendingJavaException);

	// Another version of that as an operator overload.
	// Note that you can't use this operator as an lvalue.
	const jobject	operator[](int n) const
		throw(PendingJavaException);

	/*
	 * Set the nth element of the array
	 */
	void	set(jsize n, jobject value)
		throw(PendingJavaException);

};

};	// namespace java.lang

namespace nio
{
/*
 * JNUBuffer wraps java.nio.Buffer.
 */
class JNUBuffer : public JNUObject<jobject>
{
public:
	typedef JNUBuffer	This;
	typedef JNUObject<jobject>	super;

protected:
	/*
	 * Constructor.
	 */
	JNUBuffer(const JNUObject<jobject> &object);

public:
	~JNUBuffer() throw() {}
	// Accessors

	/*
	 * Return the java.nio.Buffer, ByteBuffer, whatever it is.
	 */
	jobject			getBuffer()			{ return object(); }
	const jobject	getBuffer() const	{ return object(); }

	/*
	 * Return the capacity of the Buffer.
	 * This calls Buffer.capacity().
	 */
	virtual jsize	getCapacity() const
		throw(PendingJavaException);

	/*
	 * Return the limit of the Buffer.
	 * This calls Buffer.limit().
	 */
	virtual jsize	getLimit() const
		throw(PendingJavaException);

	/*
	 * Return the position of the Buffer.
	 * This calls Buffer.position().
	 */
	virtual jsize	getPosition() const
		throw(PendingJavaException);

	// Mutators
	/*
	 * Set the limit.
	 * This calls Buffer.limit(int).
	 */
	virtual void	setLimit(jsize limit)
		throw(PendingJavaException);

	/*
	 * Set the position.
	 * This calls Buffer.position(int)
	 */
	virtual void	setPosition(jsize position)
		throw(PendingJavaException);
};

/*
 * JNUByteBuffer wraps java.nio.ByteBuffer,
 * taking care of the differences between non-direct and direct buffers.
 */

class JNUByteBuffer : public JNUBuffer
{
public:
	typedef JNUByteBuffer	This;
	typedef JNUBuffer		super;
	/*
	 * Constructor from an existing ByteBuffer.
	 * The 'mode' parameter is zero or JNI_ABORT,
	 * as for JNUPrimitiveArray::JNUPrimitiveArray().
	 */
	JNUByteBuffer(const JNUObject<jobject> &buffer, jint mode = 0)
		 throw(PendingJavaException);

	/*
	 * Destructor.
	 */
	~JNUByteBuffer() throw();

	/*
	 * Wrap a ByteBuffer around existing data.
	 * This attemps to allocate a direct buffer,
	 * allocating a non-direct buffer only if that fails.
	 * The 'mode' parameter is zero or JNI_ABORT,
	 * as for JNUPrimitiveArray::JNUPrimitiveArray().
	 */
	static JNUByteBuffer	wrap(void *address, jsize capacity, jint mode = 0)
		 throw(PendingJavaException);

private:
	/*
	 * Construct a new JNUByteBuffer and a new NIO ByteBuffer
	 * from an existing memory address and capacity
	 */
	JNUByteBuffer(void *address, jsize capacity, jint mode = 0)
		throw(PendingJavaException);

private:
	// Private copy, assign: no copying
	JNUByteBuffer(const JNUByteBuffer &);
	JNUByteBuffer &operator=(const JNUByteBuffer&);

public:
	// Accessors

	/*
	 * Return the data
	 */
	void		*getData()					{ return _data;	}
	const void	*getData() const			{ return _data;	}

	/*
	 * Return the capacity of the ByteBuffer
	 */
	jsize	getCapacity() const
		throw(PendingJavaException)			{ return _capacity;	}

	/*
	 * Tells whether or not this byte buffer is direct.
	 */
	bool	isDirect() const				{ return _isDirect;	}

	// Mutators

	/*
	 * Commit the changes made to the data back to the ByteBuffer, with immediate effect.
	 * Note that no abort() method is provided: if the ByteBuffer is direct,
	 * the changes are already there, and if it isn't, they aren't.
	 * Nothing can be done about this problem except to know whether the buffer
	 * is direct or not.
	 */
	void	commit() throw(PendingJavaException);

private:
	void	*_data;
	jsize	_capacity;
	JNUByteArray	*_array;
	bool	_isDirect;
	static jweak	_weak;
};


};	// namespace java.nio
};	// namespace java

#ifdef _MSC_VER
#pragma warning(pop)
#endif

