// $Id: PacketCapture.java,v 1.7 2008/11/18 14:02:39 rickwall Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Copyright (C) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.capture;
import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import net.sourceforge.jpcap.net.EthernetPacket;
import net.sourceforge.jpcap.net.Packet;
import net.sourceforge.jpcap.net.PacketFactory;
import net.sourceforge.jpcap.util.Timeval;


/**
 * This class is the core of packet capture in jpcap.
 * It provides a  high-level interface for capturing network packets by encapsulating 
 * <i>libpcap</i>.
 * <p>
 * To capture network packets, implement {@link PacketListener}
 * and register with an instance of this class.
 * When packets arrive, your {@link PacketListener#packetArrived packetArrived()} callback method will be invoked.
 * Examples can be found in <code>net.sourceforge.jpcap.tutorial</code>.
 * This utilizes <i>libpcap</i>'s <code>pcap_loop(), pcap_dispatch(),</code> and <code>pcap_breakloop()</code> methods.
 * <p>
 * In addition to packet capture via callbacks, the class supports:
 * <ul>
 * <li> A simpler 'no-callbacks' packet-reading interface via {@link PacketCapture#getInputStream()}
 * (which corresponds to <code>pcap_next_ex()</code>)
 * <li> packet <i>sending</i> via {@link PacketCapture#getOutputStream()},
 * <li> packet <i>dumping</i> from the network or a <i>tcpdump</i>-formatted savefile to a <i>tcpdump</i>-formatted savefile,
 * via {@link PacketCapture#getOutputStream(File file)}
 * <li> high-efficiency packet <i>filtering</i> from the network or a <i>tcpdump</i>-formatted savefile to a <i>tcpdump</i>-formatted savefile,
 * via {@link PacketCapture#captureTo(int, PacketOutputStream)}.
 * </ul>
 * So you can use this class:
 * <ol>
 * <li>to capture and filter packets
 * <li>to save packets to savefiles
 * <li>to filter a live capture directly to an output savefile, or
 * <li>to filter an input savefile directly to an output savefile.
 * </ol>
 * There are several parameters that control packet capture:
 *<dl>
 *<dt><code>timeout</code>
 *<dd>This is specified when you open the capture device, which defaults to 1 second.
 * The timeout allows the kernel layer to buffer up several packets for transfer to the application at the same time, which is much more efficient than transferring one at a time (fewer context switches).
 * You have to trade that off against your application's requirement for timeliness.
 *<dt>{@link PacketCapture.LoopMode}
 *<dd>This is set by {@link PacketCapture#setLoopMode setLoopMode()}, which defaults
 * to {@link PacketCapture.LoopMode#Loop Loop}.
 *<dt>'minimum kernel copy' value
 *<dd>On Winpcap platforms, this is as documented for
 * {@link JWinPcap#getMinimumKernelCopy} and {@link JWinPcap#setMinimumKernelCopy setMinimumKernelCopy()};
 * on non-Winpcap platforms it defaults to an unknown but probably similar value.
 *<dt>capture method
 *<dd>The capture method you call can be one of {@link PacketCapture#capture capture()},
 * {@link PacketCapture#captureTo captureTo()},
 * {@link PacketCapture#captureNext captureNext()}, or
 * {@link PacketInputStream#readPacket}.
 *<dt><code>count</code>
 *<dd>The count supplied to {@link PacketCapture#capture capture()}.
 *</dl>
 * The operation of these is explained below.
 * <ol>
 * <li>{@link PacketCapture#capture capture()} in {@link PacketCapture.LoopMode#Loop Loop} mode
 * loops doing the following until the specified count is reached:
 * <ol type="a">
 * <li>buffer packets until the 'minimum kernel copy' value is reached,
 * or the timeout expires, (whichever occurs first), then
 * <li>call your callback with the buffered packets one by one.
 * </ol>
 * <li>{@link PacketCapture#capture capture()} in {@link PacketCapture.LoopMode#Dispatch Dispatch} mode
 * does the same steps, then resumes looping, but, on any loop iteration,
 * if the timeout expires without any packets being captured, it returns.
 * <p>
 * Neither of these throws {@link java.net.SocketTimeoutException}.
 * If a timeout occurs, Loop mode just keeps running until the count expires;
 * Dispatch mode returns immediately.
 * <p>
 * You can see how many packets were captured via {@link PacketCapture#getStatistics}.
 * <li>{@link PacketCapture#captureTo captureTo()} behaves the same as capture(),
 * i.e. it is subject to the LoopMode, MinimumKernelCopy, and timeout settings.
 * <li>{@link PacketCapture#captureNext} captures and returns <i>one</i> packet,
 * without a callback, but it is still subject to the timeout specified when opening.
 * If the timeout expires it throws {@link java.net.SocketTimeoutException}.
 * <li>{@link PacketInputStream#readPacket} is identical to captureNext() under the hood.
 * </ol>
 *<p>
 * <b>Note: <i>Version 2.0.0 binary incompatibility:</i></b>
 * All the native methods in this class have been renamed, are now private, and in almost all cases
 * now have different calling sequences.
 * The old public native methods of this class that took <code>instance</code> arguments:
 * <ul>
 * <li>were made public unintentionally
 * <li>were unsafe, unchecked, and liable to cause JVM crashes
 * <li>are now unimplementable, because the concept of 'instance numbers' has been banished, and therefore
 * <li><i>have been removed</i>.
 * </ul>
 * Any code that calls these methods it must be changed.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.7 $
 * @lastModifiedBy $Author: rickwall $
 * @lastModifiedAt $Date: 2008/11/18 14:02:39 $
 * @see PacketCaptureCapable
 * @see PacketInputStream
 * @see PacketOutputStream
 * @see PacketListener
 */
public class PacketCapture extends PacketCaptureBase
  implements PacketCaptureCapable
{
	/**
	 * Capture Direction enumeration.
	 * This is used in the {@link #setDirection} method
	 * to determine the 'direction' of packet capture,
	 * i.e. whether to capture packets coming in, going out, or both.
	 * The values of these correspond with the same items in pcap.h.
	 *
	 * @since 2.0.0
	 */
	public enum Direction
	{
		/**
		 * PCAP_D_INOUT will capture packets received by or sent by the device.
		 */
		PCAP_D_INOUT,
		/**
		 * PCAP_D_IN will only capture packets received by the device.
		 */
		PCAP_D_IN,
		/**
		 * PCAP_D_OUT will only capture packets sent by the device.
		 */
		PCAP_D_OUT
	};
	
	/**
	 * LoopModes determine how the {@link #capture} and {@link #captureTo} methods behave in the presence of timeouts.
	 * For <i>libpcap</i> programmers, this enumeration specifies whether to call <i>pcap_loop()</i>
	 * or <i>pcap_dispatch()</i>.
	 *<p>
	 * If capturing from a file, LoopMode.Dispatch is identical to LoopMode.Loop.
	 *<p>
	 * If not set, the default LoopMode is LoopMode.Loop.
	 *
	 * @see #capture
	 * @see #getLoopMode
	 * @see #setLoopMode
	 * @see #configureBlocking
	 *
	 * @since 2.0.0
	 */
	public enum LoopMode
	{
		/**
		 * The {@link #Loop} mode causes packet captures to run until either:
		 *<ul>
		 * <li>the count is exhausted (if positive),
		 * <li>the input file is exhausted (if capturing from a savefile), or
		 * <li>{@link #endCapture} is called.
		 *</ul>
		 * The <code>timeout</code> parameter specified when opening the capture device is used
		 * only to determine how long the capture loop waits for another packet to arrive before
		 * starting to call callbacks for those that have already arrived: this is a measure to
		 * reduce context switches.
		 *<p>
		 * For <i>libpcap</i> programmers, this value causes the capture methods to call <i>pcap_loop()</i>.
		 */
		Loop,
		/**
		 * The {@link #Dispatch} mode causes packet captures to run until either:
		 * <ul>
		 * <li>no packets have arrived within the timeout period specified when the capture device was opened,
		 * <li>the count is exhausted (if positive),
		 * <li>the input file is exhausted (if capturing from a savefile), or
		 * <li>{@link #endCapture} is called.
		 *</ul>
		 * <p>In addition, if the capture device is in non-blocking mode, the capture methods return immediately
		 * if there are no packets ready to be captured at the moment it is called.
		 * <p>
		 * The <code>timeout</code> parameter specified when opening the capture device is used
		 * both for this purpose <i>and</i> to determine how long the capture loop waits for another packet to arrive before
		 * starting to call callbacks for those that have already arrived: this is a measure to
		 * reduce context switches.
		 * <p>
		 * For <i>libpcap</i> programmers, this value causes the capture methods to call <i>pcap_dispatch()</i>.
		 *
		 * @see #configureBlocking
		 */
		Dispatch;
	}
	
	/**
	 * The current loop mode.
	 * @since 2.0.0
	 */
	private LoopMode	loopMode = LoopMode.Loop;
	
	/**
	 * These are just here to cause the corresponding JNI constants to be generated.
	 * Sadly that doesn't happen if we assign the ordinal values directly.
	 */
	static final int	LOOP_MODE = 0,	// LoopMode.Loop.ordinal(),
		DISPATCH_MODE = 1;	// LoopMode.Dispatch.ordinal();
	
	// ... so we add a sanity check.
	static
	{
		assert LOOP_MODE == LoopMode.Loop.ordinal() : "LoopMode.Loop ordinal is broken";
		assert DISPATCH_MODE == LoopMode.Dispatch.ordinal() : "LoopMode.Dispatch ordinal is broken";
	}
	
/**
	 * Controls the initial value of VERBOSE in the JNI, if you rerun javah and recompile it.
	 * @since 2.0.0
	 */
	private static final int VERBOSE = 0;
	/**
	 * Name of the java-enabled wrapper around <i>libpcap</i>.
	 */
	public static final String LIB_PCAP_WRAPPER = "jpcap";
	/**
	 * Maximum size of a capture buffer, snaplen, Ethernet packet, etc.
	 */
	public static final int	MAX_SNAPLEN = 0x600;

  // Instance data.
  /**
   * the JNI handle of this instance
   * @since 2.0.0
   */
  long handle;
  /**
   * Open/closed state.
   * @since 2.0.0
   */
  boolean	isClosed = true;
  private PacketInputStream		inputStream;
  PacketOutputStream	networkOutputStream;
  PacketOutputStream	saveOutputStream;
  /**
   * The current capture device.
   * @since 2.0.0
   */
  private CaptureDevice	device;
  /**
   * Snaplen specified when this device was opened.
   * @since 2.0.0
   */
  int	snaplen = MAX_SNAPLEN;
 
public int getSnaplen() {
	return snaplen;
}

public void setSnaplen(int snaplen) {
	this.snaplen = snaplen;
}

/**
   * Timeout specified when this device was opened.
   * @since 2.0.0
   */
  int	timeout = DEFAULT_TIMEOUT;
  /**
   * Netmask of the currently open capture device,
   * for {@link #compileFilter} and {@link #setFilter() setFilter()};
   * ultimately for pcap_compile() in the JNI. If provided as non-zero,
   * allows filtering on IPv4 broadcasts.
   * @see #setFilter(CompiledFilter)
   * @since 2.0.0
   */
  int	netmask;
  /**
   * Remember the input device/savefile when opening.
   * @since 2.0.0
   */
  File		inputFile;
  /**
   * Current blocking mode.
   * @since 2.0.0
   */
  boolean	blocking = true;

  /**
   * Create a new packet capture instance.
   * Allocates an 'instance number' from zero to INSTANCE_MAX.
   * Only INSTANCE_MAX instances of this class can exist concurrently.
   * This is tied to declarations in the JNI code.
   */
  public PacketCapture()
  {
	  this.handle = create0();
  }
  
  /**
   * Create the JNI instance.
   */
  private native long	create0();

  /**
   * Construct a new instance and open the specified device for capture.
   * This is equivalent to:
   *<pre>
   *<code>
   *	PacketCapture packetCapture = new PacketCapture();
   *	packetCapture.open(device, DEFAULT_SNAPLEN, promiscuous, DEFAULT_TIMEOUT);
   *<code>
   *</pre>
   * @param device CaptureDevice to capture from.
   * @param promiscuous whether or not the device should be put into 'promiscuous' mode.
   * @exception CaptureDeviceOpenException can't open the specified device
   *
   * @since 2.0.0
   */
  public PacketCapture(CaptureDevice device, boolean promiscuous)
   throws CaptureDeviceOpenException
  {
	  this(device, DEFAULT_SNAPLEN, promiscuous, DEFAULT_TIMEOUT);
  }

  /**
   * Construct a new instance and open the specified device for capture.
   * This is equivalent to:
   *<pre>
   *<code>
   *	PacketCapture packetCapture = new PacketCapture();
   *	packetCapture.open(device, snaplen, promiscuous, timeout);
   *<code>
   *</pre>
   * @param snaplen the 'snapshot' length: defines the maximum number of 
   * bytes to save from each captured packet.
   * @param promiscuous whether or not the device should be put into
   * 'promiscuous' mode: true => capture all packets; false => capture packets only to and from this system.
   * @param timeout the packet capture timeout in milliseconds.
   * Defines the maximum amount of time that capture will wait for another packet,
   * if the capture device's buffer isn't full,
   * before notifying the application of the packets already captured.
   * A value of zero means return each packet immediately: this mode is rather inefficient.
   * Timeouts of 500ms-2 seconds are reasonable, or longer if the application can wait longer.
   * @exception CaptureDeviceOpenException can't open the specified device
   *
   * @since 2.0.0
   */
  public PacketCapture(CaptureDevice device, int snaplen, boolean promiscuous, int timeout)
   throws CaptureDeviceOpenException
  {
	  this();
	  open(device, snaplen, promiscuous, timeout);
  }

  /**
   * Construct a new instance and open the specified file for capture.
   * This is equivalent to:
   *<pre>
   *<code>
   *	PacketCapture packetCapture = new PacketCapture();
   *	packetCapture.openOffline(file.getPath());
   *<code>
   *</pre>
   * @param file <i>tcpdump</i>-formatted savefile to capture from.
   * @exception CaptureFileOpenException can't open capture file
   *
   * @since 2.0.0
   */
  public PacketCapture(File file)
   throws CaptureFileOpenException
  {
	  this();
	  openOffline(file.getPath());
  }
  
  /**
   * {@inheritDoc}
   *<p>
   * Releases the JNI 'instance' number so it can be reused.
   *
   * @since 2.0.0
   */
  protected void	finalize() throws Throwable
  {
	  if (handle != 0)
		  destroy0(handle);
  }
  
  private native void	destroy0(long handle);
  
  /**
   * Compile a filter expression for the current capture device. 
   *
   * @param filter Filter expression.
   * @param optimize Controls whether the filter is optimized when compiled.
   * @return a CompiledFilter, for use with {@link #setFilter(CompiledFilter)}.
   * @exception InvalidFilterException Invalid filter syntax, or capture device is not open.
   */
  public CompiledFilter	compileFilter(String filter, boolean optimize) throws InvalidFilterException
  {
	  if (isClosed())
		  throw new InvalidFilterException("capture device not open");
//	  System.out.println("compileFilter: "+filter);
	  return new CompiledFilterImpl(filter, optimize, netmask);
  }
  
  /**
   * Configure the blocking mode.
   * Non-blocking mode causes the capture methods to return immediately if
   * there are no packets to capture, rather than waiting for the specified timeout period.
   * When capturing from a savefile, this method has no effect.
   *
   * @param blocking If true, this channel will be placed in blocking mode;
   * if false, it will be placed non-blocking mode.
   * @exception IOException I/O problem changing the blocking mode.
   * @since 2.0.0
   */
  public synchronized void	configureBlocking(boolean blocking) throws IOException
  {
	  if (isClosed())
		  throw new IOException("capture device not open");
	  configureBlocking0(handle, blocking);
	  this.blocking = blocking;
  }
  
  // native support
  private native void	configureBlocking0(long handle, boolean isBlocking) throws IOException;
 
  /**
   * Return true if the capture device is in blocking mode, false for non-blocking mode.
   * @return true iff this channel is in blocking mode.
   * @exception IOException I/O problem determining the blocking mode (e.g. capture device not open).
   * @since 2.0.0
   */
  public boolean	isBlocking() throws IOException
  {
	  if (isClosed())
		  throw new IOException("capture device not open");
	  // NB blocking mode is saved here to avoid having to define a native method.
	  return blocking;
  }

  // methods for controlling a packet capture session

  /**
   * Open a network device for data capture.
   * This is equivalent to <code>open(device, DEFAULT_SNAPLEN, promiscuous, DEFAULT_TIMEOUT);</code>
   * 
   * @param device the CaptureDevice to open. 
   * @param promiscuous whether or not the device should be put into
   * 'promiscuous' mode: true => capture all packets; false => capture packets only to and from this system.
   *
   * @exception CaptureDeviceOpenException can't open the specified device
   *
   * @see #open(CaptureDevice, int, boolean, int)
   *
   * @since 2.0.0
   */
  public void	open(CaptureDevice device, boolean promiscuous)
  throws CaptureDeviceOpenException
  {
	  open(device, DEFAULT_SNAPLEN, promiscuous, DEFAULT_TIMEOUT);
  }

  /**
   * Open a network device for data capture.
   * 
   * @param device the CaptureDevice to open. 
   * @param snaplen the 'snapshot' length. Defines the maximum number of 
   * bytes to save from each captured packet.
   * @param promiscuous whether or not the device should be put into
   * 'promiscuous' mode: true => capture all packets; false => capture packets only to and from this system.
   * @param timeout the packet capture timeout in milliseconds.
   * Defines the maximum amount of time that capture will wait for another packet,
   * if the capture device's buffer isn't full,
   * before notifying the application of the packets already captured.
   * A value of zero means return each packet immediately: this mode is rather inefficient.
   * Timeouts of 500ms-2 seconds are reasonable, or longer if the application can wait longer.
   * @exception CaptureDeviceOpenException can't open the specified device
   *
   * @since 2.0.0
   */
  public void	open(CaptureDevice device, int snaplen, boolean promiscuous, int timeout)
  throws CaptureDeviceOpenException
  {
	  if (!isClosed())
		  throw new CaptureDeviceOpenException("already open");
	  if (device == null)
		  throw new NullPointerException("device is null");
	  super.linkType = open0(handle, device.getName(), snaplen, promiscuous, timeout);
	  this.device = device;
	  this.snaplen = snaplen;
	  this.timeout = timeout;
	  this.netmask = device.getNetmask();
	  isClosed = false;
  }
  
  /** {@inheritDoc} */
  public void open(String device, boolean promiscuous)
   throws CaptureDeviceOpenException {
    open(device, DEFAULT_SNAPLEN, promiscuous, DEFAULT_TIMEOUT);
  }

  /** {@inheritDoc} */
  public void open(String device, int snaplen, 
                   boolean promiscuous, int timeout) 
    throws CaptureDeviceOpenException {
	  if (!isClosed())
		  throw new CaptureDeviceOpenException("already open");
	  if (device == null)
		  throw new NullPointerException("device name is null");
	  super.linkType = open0(handle, device, snaplen, promiscuous, timeout);
	  this.snaplen = snaplen;
	  this.timeout = timeout;
	  try
	  {
		  this.netmask = getNetmask(device);
	  }
	  catch (CaptureConfigurationException exc)
	  {
		  // ignored, bad luck, no netmask, no filtering on IPv4 broadcast addresses.
		  this.netmask = 0;
	  }
	  isClosed = false;
  }

  /**
   * Native open method.
   */
  private native int open0(long handle, String device, int snaplen, 
                          boolean promiscuous, int timeout) 
    throws CaptureDeviceOpenException;


  /** {@inheritDoc} */
  public void openOffline(String fileName) throws CaptureFileOpenException {
	  if (!isClosed())
		  throw new CaptureFileOpenException("already open");
	  if (fileName == null)
		  throw new NullPointerException("file name is null");
	  super.linkType = openOffline0(handle, fileName);
	  this.inputFile = new File(fileName);
	  isClosed = false;
  }

  /**
   * Native openOffline method.
   */
  private native int openOffline0(long handle, String fileName)
    throws CaptureFileOpenException;
 
  /**
   * {@inheritDoc}
   *<p>
   * This is a shorthand for <code>setFilter(compileFilter(filterExpression, optimize));</code>
   */
  public void setFilter(String filterExpression, boolean optimize)
    throws InvalidFilterException {
	  if (isClosed())
		  throw new InvalidFilterException("capture device is not open");
	  setFilter(compileFilter(filterExpression, optimize));
  }

  /**
   * Set the filter from a precompiled filter.
   * @param filter CompiledFilter, from {@link #compileFilter}
   * @exception InvalidFilterException Capture device is not open
   */
  public void	setFilter(CompiledFilter filter) throws InvalidFilterException
  {
	  if (isClosed())
		  throw new InvalidFilterException("capture device is not open");
//	  System.out.println("setFilter: "+filter);
	  setFilter0(handle, filter.getHandle());
  }

  /**
   * Native setFilter method.
   */
  private native void setFilter0(long handle, long filterHandle)
    throws InvalidFilterException;

  /** {@inheritDoc} */
  public void capture(int count) throws CapturePacketException {
	  if (isClosed())
		  throw new CapturePacketException("capture device closed");
	  try
	  {
		  int	rc = capture0(handle, count, loopMode.ordinal(), timeout);
	  }
	  catch (EOFException exc)
	  {
		  ;
	  }
	  catch (IOException exc)
	  {
		  throw new CapturePacketException(exc);
	  }
  }

  /**
   * Native capture method.
   * @param handle Instance
   * @param count Number of packets to capture, or -1 meaning infinity.
   * @param mode Capture mode
   * @param timeout Timeout specified when the capture device was opened.
   * @return the number of packets captured; 0 meaning EOF, or -2 meaning endCapture() called before any packets captured.
   * @exception CapturePacketException some <i>libpcap</i> error
   * @exception EOFException EOF on savefile
   */
  private native int capture0(long handle, int count, int mode, int timeout)
    throws CapturePacketException, EOFException;

  /**
   * Return the next packet captured.
   * For <i>libpcap</i> programmers, this method calls <i>pcap_next_ex()</i>.
   * @return a Packet
   * @exception CapturePacketException some <i>libpcap</i> error.
   * @exception SocketTimeoutException The timeout specified when the capture device was opened has expired without a packet being captured.
   * @exception EOFException EOF on savefile.
   * @exception IOException capture device is closed.
   */
  public Packet	captureNext()
	throws CapturePacketException,
	  SocketTimeoutException,
	  EOFException,
	  IOException
  {
	  if (isClosed)
		  throw new IOException("capture device is closed");
	  // NB the JNI code requires this to be allocated via allocateDirect()
	  ByteBuffer	bb = ByteBuffer.allocateDirect(snaplen);
	  //ByteBuffer	bb = ByteBuffer.allocate(snaplen);
	  try
	  {
		  long	ts = captureNext0(handle, bb);
	//	  System.out.println("captureNext: ts="+ts+" bb.limit()="+bb.limit());
		  Timeval	tv = new Timeval((int)(ts/1000000), (int)(ts & 1000000));
		  Packet	packet = PacketFactory.dataToPacket(linkType, bb, tv);
		  return packet;
	  }
//	  catch (EOFException exc)
//	  {
//		  System.out.println("captureNext(): caught EOFException "+exc);
//		  throw exc;
//	  }
	  finally
	  {
		  
	  }
  }
  
  /**
   * Native 'next-packet' method.
   * @param handle Instance
   * @param bb Direct ByteBuffer to return packet data in: allocating this once rather than per packet is
   * a <i>significant</i> efficiency.
   * @return the timestamp of the captured packet in microseconds
   * @exception CapturePacketException some <i>libpcap</i> error
   * @exception EOFException EOF on savefile
   */
  private native long	captureNext0(long handle, ByteBuffer bb) throws CapturePacketException, EOFException;
  
  /**
   * Capture packets and dump them to the network or a <i>tcpdump</i>-formatted savefile.
   * You can do this manually yourself but this method is considerably
   * more efficient, as it doesn't have to construct Packet objects or
   * cross and re-cross the Java/JNI interface.
   *<p>Note that to replay a savefile, you need to call this method on a PacketCapture handle
   * which is open 'offline', supplying a PacketOutputStream created by {@link PacketCapture#getOutputStream()}
   * from <i>another</i> PacketCapture handle that was created to capture 'online', i.e. with an open network device.
   *
   * @param count Maximum count of packets to capture, or -1 for infinity.
   * @param out PacketOutputStream to dump to.
   * @return the number of packets captured and dumped.
   * @exception CapturePacketException Packet capture problem.
   * @exception IOException I/O problem on dump file.
   * @exception IllegalArgumentException if this PacketCapture object is capturing <i>from</i> the network
   * and the output stream is dumping <i>to</i> the network, as this will just break your network.
   *
   * @since 2.0.0
   */
  public int	captureTo(int count, PacketOutputStream out) throws CapturePacketException, IOException, IllegalArgumentException
  {
	  return out.captureFrom(this, count);
  }

  int	captureToFile(int count) throws CapturePacketException, IOException
  {
	  return captureToFile0(handle, count);
  }
  
  /**
   * Although it could be implemented at the Java level, this method is implemented entirely at the JNI level, which is much simpler and a lot more efficient.
   */
  native int captureToFile0(long handle, int count) throws CapturePacketException, IOException;

  int	captureToNetwork(int count, long outHandle) throws CapturePacketException, IOException
  {
	  return captureToNetwork0(handle, count, outHandle);
  }
  
  /**
   * Although it could be implemented at the Java level, this method is implemented entirely at the JNI level, which is much simpler and a lot more efficient.
   * @param handle JNI handle
   * @param count Count of packets to capture, or -1 for infinity.
   * @param outHandle JNI handle of output device.
   */
  native int captureToNetwork0(long handle, int count, long outHandle) throws CapturePacketException, IOException;

  /** {@inheritDoc} */
  public CaptureStatistics getStatistics() {
	  return getStatistics0(handle);
  }

  /**
   * The packet capture library sets up the statistic counter members
   * when this method is invoked internally.
   */
  private native CaptureStatistics getStatistics0(long handle);

  public synchronized void close()
  {
	  try
	  {
		  if (!isClosed)
		  {
			  if (inputStream != null)
				  inputStream.close();
			  close0(handle);
		  }
	  }
	  catch (IOException exc)
	  {
		  throw new RuntimeException("unexpected IOException in close: "+exc.getMessage());
	  }
	  finally
	  {
		  device = null;
		  inputStream = null;
		  networkOutputStream = null;
		  inputFile = null;
		  isClosed = true;
	  }
  }

  /**
   * Native close method.
   */
  private native void close0(long handle);

  /**
   * Return true if the captor is closed or a device or capture file hasn't been opened yet.
   * @return false iff the captor is currently open.
   */
  public boolean	isClosed()
  {
	  return isClosed;
  }
  
  /**
   * Return the current capture device.
   * @return the current capture device, or null if not opened via a CaptureDevice, or closed.
   * @since 2.0.0
   */
  public CaptureDevice	getDevice()
  {
	  return device;
  }

  /** {@inheritDoc} */
  public void endCapture() {
	  if (isClosed)
		  throw new IllegalStateException("Capture is closed");
    endCapture0(handle);
  }

  /**
   * Native endCapture method.
   */
  private native void endCapture0(long handle);

  /**
   * Find a suitable CaptureDevice for live capture.
   * This just returns the first element returned by {@link #lookupCaptureDevices()}
   * that isn't a loopback device.
   *
   * @return a CaptureDevice.
   * @exception CaptureDeviceNotFoundException No usable capture device found
   * @exception CaptureConfigurationException Capture device configuration problem
   *
   * @since 2.0.0
   */
  public static CaptureDevice	findCaptureDevice()
    throws CaptureDeviceNotFoundException, CaptureConfigurationException
  {
	  try
	  {
		  for (CaptureDevice device : lookupCaptureDevices())
			  if (
			  !device.isLoopback()
			  && device.getNetwork() != 0
			  )
				  return device;
	  }
	  catch (CaptureDeviceLookupException exc)
	  {
		  throw new CaptureDeviceNotFoundException(exc);
	  }
	  throw new CaptureDeviceNotFoundException("no usable capture devices detected");
  }

  /**
   * {@inheritDoc}
   *
   * @deprecated Please use {@link PacketCapture#findCaptureDevice}.
   */
  public String findDevice()
  throws CaptureDeviceNotFoundException
  {
	  // return findDevice0();
	  try
	  {
		  return findCaptureDevice().getName();
	  }
	  catch (CaptureConfigurationException exc)
	  {
		  throw new CaptureDeviceNotFoundException(exc);
	  }
  }

  // Native support.
//  private native String findDevice0() throws CaptureDeviceNotFoundException;
  
  /**
   * Get the interface list as an array of <code>CaptureDevice[]</code>.
   * The JNI implementation under this method uses standard <i>libpcap</i> code,
   * not system-dependent code, so it should be much less fragile at the Jpcap.dll level
   * than {@link #lookupDevices} used to be.
   * <p>
   * As it returns {@link CaptureDevice CaptureDevices}, which have complete device information in them,
   * including a <i>list</i> of IPv4 or IPv6 address/netmask information,
   * this method also obsoletes the
   * {@link #getNetwork} and {@link #getNetmask} methods. 
   *
   * @return an array of {@link CaptureDevice CaptureDevices}, possibly zero-length, never null.
   * Devices that aren't 'up' or don't have IP addresses are not returned.
   * @exception CaptureDeviceLookupException Capture device lookup problem
   * @exception CaptureConfigurationException Capture device configuration problem
   *
   * @since 2.0.0
   */
  public static CaptureDevice[]	lookupCaptureDevices()
    throws CaptureDeviceLookupException, CaptureConfigurationException
  {
	  try
	  {
		  CaptureDevice[]	devices = lookupCaptureDevices0();
		  return devices;
	  }
	  catch (ClassNotFoundException exc)
	  {
		  throw new CaptureDeviceLookupException(exc);
	  }
	  catch (NoSuchMethodException exc)
	  {
		  throw new CaptureDeviceLookupException(exc);
	  }
	  catch (NoSuchFieldException exc)
	  {
		  throw new CaptureDeviceLookupException(exc);
	  }
  }
  
  private static native CaptureDevice[] lookupCaptureDevices0()
    throws CaptureDeviceLookupException, CaptureConfigurationException,
	  ClassNotFoundException, NoSuchMethodException, NoSuchFieldException;
  
  /**
   * Get Interface List.
   * <p>
   * This API will be removed in a future release.
   *
   * @deprecated Please use {@link #lookupCaptureDevices}.
   * @return Network device interface names.
   * @exception CaptureDeviceLookupException Capture device lookup problem
   */
  public static String[] lookupDevices() 
    throws CaptureDeviceLookupException
  {
	  // return lookupDevices0();
	  try
	  {
		  CaptureDevice[]	devices = lookupCaptureDevices();
		  String[]	result = new String[devices.length];
		  for (int i = 0; i < devices.length; i++)
			  result[i] = devices[i].getName();
		  return result;
	  }
	  catch (CaptureConfigurationException exc)
	  {
		  throw new CaptureDeviceLookupException(exc);
	  }
  }

//  private static native String[] lookupDevices0() 
//    throws CaptureDeviceLookupException;

  // the following methods could be static, but aren't so that they 
  // can be included in the PacketCaptureCapable interface.
  
  /**
   * {@inheritDoc}
   *
   * @deprecated Please use {@link #lookupCaptureDevices}.
   */
  public int	getNetwork(String name)
  throws CaptureConfigurationException
  {
	  try
	  {
		  CaptureDevice[]	devices = lookupCaptureDevices();
		  for (CaptureDevice device : devices)
			  if (device.getName().equals(name))
				  return device.getNetwork();
	  }
	  catch (CaptureDeviceLookupException exc)
	  {
		  throw new CaptureConfigurationException(exc);
	  }
	  throw new CaptureConfigurationException("device not found: "+name);
  }
  
  // public native int	getNetwork(String device) throws CaptureConfigurationException;
  
  /**
   * {@inheritDoc}
   *
   * @deprecated Please use {@link #lookupCaptureDevices}.
   */
  public int	getNetmask(String name)
  throws CaptureConfigurationException
  {
	  try
	  {
		  CaptureDevice[]	devices = lookupCaptureDevices();
		  for (CaptureDevice device : devices)
			  if (device.getName().equals(name))
				  return device.getNetmask();
	  }
	  catch (CaptureDeviceLookupException exc)
	  {
		  throw new CaptureConfigurationException(exc);
	  }
	  throw new CaptureConfigurationException("device not found: "+name);
  }
  
  // public native int	getNetmask(String device) throws CaptureConfigurationException;
  
  public String	getLinkLayerDescription() throws CaptureConfigurationException
  {
	  return getLinkLayerDescription(linkType);
  }

  /**
   * Return the link layer description for the specified type.
   * @param linkType Link type.
   * @return the link layer description.
   * @exception CaptureConfigurationException I/O problem getting description.
   * @since 2.0.0
   */
  public String	getLinkLayerDescription(int linkType) throws CaptureConfigurationException
  {
	  String	result = getLinkLayerDescription0(linkType);
	  if (result == null)
		  throw new CaptureConfigurationException("unknown link type "+linkType);
	  return result;
  }
  
  // native support
  private native String	getLinkLayerDescription0(int type) throws CaptureConfigurationException;

  /**
   * Return the link layer name, e.g. "eth10".
   * @return the link layer name.
   * @exception CaptureConfigurationException Error finding the name
   * @since 2.0.0
   */
  public String	getLinkLayerName() throws CaptureConfigurationException
  {
	  return getLinkLayerName0(linkType);
  }
  
  /**
   * Return the link layer name for the specified linkType.
   * @param linkType link type.
   * @return the link layer name.
   * @exception CaptureConfigurationException Error finding the name
   * @since 2.0.0
   */
  public String	getLinkLayerName(int linkType) throws CaptureConfigurationException
  {
	  return getLinkLayerName0(linkType);
  }
  
  // native support
  private native String	getLinkLayerName0(int linkType) throws CaptureConfigurationException;

  /** {@inheritDoc} */
  public int getLinkLayerType() throws CaptureConfigurationException {
//    return getLinkLayerType0(handle);
	  return linkType;
  }

  private native int getLinkLayerType0(long handle)
    throws CaptureConfigurationException;
  
  /**
   * Get the available link layer types for the current capture device.
   * @return an array of link layer types, suitable for use with {@link #setLinkLayerType}.
   * 
   * @exception CaptureConfigurationException Problem getting the link layer types.
   *
   * @since 2.0.0
   */
  public int[]	getLinkLayerTypes() throws CaptureConfigurationException
  {
	  if (isClosed())
		  throw new CaptureConfigurationException("capture device closed");
	  return getLinkLayerTypes0(handle);
  }
  
  // native support
  private native int[]	getLinkLayerTypes0(long handle) throws CaptureConfigurationException;
  
  /**
   * Set the current link layer type.
   * @param linkType Link layer type: must be a supported type for the
   * current adapter, e.g. as returned by {@link #getLinkLayerTypes}.
   * 
   * @exception CaptureConfigurationException Problem setting the link layer type.
   *
   * @since 2.0.0
   */
  public void	setLinkLayerType(int linkType) throws CaptureConfigurationException
  {
	  if (isClosed())
		  throw new CaptureConfigurationException("capture device closed");
	  setLinkLayerType0(handle, linkType);
	  this.linkType = linkType;
  }
  
  // native support
  private native void	setLinkLayerType0(long handle, int linkType) throws CaptureConfigurationException;

  /** {@inheritDoc} */
  public int getSnapshotLength() {
	  return isClosed? -1 : snaplen;
  }
  
  // private native int getSnapshotLength0();
  
  /**
   * Return a string giving information about the version of the <i>libpcap</i> library being used;
   * note that it contains more information than just a version number.
   * @return String.
   *
   * @since 2.0.0
   */
  public static String	getLibVersion()
  {
	  return getLibVersion0();
  }
  
  // Native support for getLibVersion()
  private static native String	getLibVersion0();
  
  /**
   * Get the current loop mode.
   * @return the current loop mode.
   * @since 2.0.0
   */
  public LoopMode	getLoopMode()
  {
	  return loopMode;
  }
  
  /**
   * Set the current loop mode.
   *<p>
   * Calling this method while a capture is in progress may or may not alter the mode.
   *
   * @param loopMode New loop mode.
   * @since 2.0.0
   */
  public void	setLoopMode(LoopMode loopMode)
  {
	  setLoopMode0(handle, loopMode.ordinal());
	  this.loopMode = loopMode;
  }
  
  // native support
  // At the JNI level, invalid loopMode values are just ignored.
  private native void	setLoopMode0(long handle, int loopMode);
  
  // Input/output stream methods
  
  /**
   * Return a PacketInputStream for reading packets from the network.
   * Multiple calls will always return the same input stream while the same
   * capture device is open.
   *
   * @return a PacketInputStream.
   * @exception IOException I/O problem creating the stream.
   * @since 2.0.0
   */
  public synchronized PacketInputStream	getInputStream()
  throws IOException
  {
	  if (inputStream == null)
		  inputStream = new CaptureInputStream();
	  return inputStream;
  }
  
  /**
   * Return a PacketOutputStream that sends packets to the network
   * device opened by this PacketCapture.
   * Repeated calls against the same captor return the same networkOutputStream.
   * 
   * @return a packet networkOutputStream.
   * @exception IOException I/O problem, e.g. the capture device or file is closed.
   *
   * @since 2.0.0
   */
  public synchronized PacketOutputStream	getOutputStream() throws IOException
  {
	  if (isClosed())
		  throw new IOException("capture device closed");
	  if (networkOutputStream == null)
	  {
		  networkOutputStream = new NetworkOutputStream(handle);
	  }
	  return networkOutputStream;
  }

  /**
   * Return a PacketOutputStream that dumps packets to a <i>tcpdump</i>-formatted savefile.
   * Only one packet saveOutputStream stream can be open at a time.
   * 
   * @param file File to dump packets to, in <i>tcpdump</i> format.
   * 
   * @return a packet writer for this captor.
   * @exception IOException I/O problem opening the dump file, or a packet saveOutputStream is currently open.
   * @since 2.0.0
   */
  public synchronized PacketOutputStream	getOutputStream(int linkType, int snaplen, File file)
	throws IOException
  {
	  if (saveOutputStream != null)
		  throw new IOException("a packet dumper is already open");
	  saveOutputStream = new TCPDumpOutputStream(linkType, snaplen, file);
	  return saveOutputStream;
  }

  /**
   * Set the direction of packet capture.
   * {@link Direction#PCAP_D_INOUT} is the default setting if this function is not called.
   * This isn't necessarily supported on all platforms;
   * some platforms might throw <code>UnsupportedOperationException</code>,
   * and some other platforms might not support {@link Direction#PCAP_D_OUT}.
   * This operation is not supported if a 'savefile' is being read.
   *
   * <b>Note:</b> not supported on Windows.
   *
   * @param direction {@link Direction} of packet capture.
   * @exception IOException Problem setting the capture direction.
   * @exception UnsupportedOperationException Operation not supported on this platform
   *
   * @see Direction
   *
   * @since 2.0.0
   */
  public void	setDirection(Direction direction) throws IOException, UnsupportedOperationException
  {
	  if (isClosed)
		  throw new IOException("capture device not open");
	  setDirection0(handle, direction.ordinal());
  }

  // Native support for setDirection()
  private native void	setDirection0(long handle, int direction) throws IOException, UnsupportedOperationException;

  /**
   * CaptureInputStream
   * @since 2.0.0
   */
  class CaptureInputStream extends PacketInputStream
  {
	  boolean	isClosed = false;
	  
	  /**
	   * @inheritDoc}
	   *<p>
	   * Close the input stream. This stops the associated capture process completely.
	   */
	  @Override
	  public synchronized void	close() throws IOException
	  {
		  if (!isClosed)
		  {
			  isClosed = true;
			  // End the capture.
			  endCapture();
		  }
	  }
	  
	  @Override
		  public synchronized Packet	readPacket() throws IOException
	  {
		  if (isClosed)
			  throw new IOException("stream is closed");
		  try
		  {
			  return captureNext();
		  }
		  catch (CapturePacketException exc)
		  {
			  throw new IOException(exc.getMessage());
		  }
	  }
  }
  
  /**
   * NetworkOutputStream class
   * @since 2.0.0
   */
  class NetworkOutputStream extends PacketOutputStream
  {
	  // Save the handle which created me,
	  // so that *another* PacketCapture can use me as an output,
	  // i.e. a PacketCapture capturing offline from a savefile.
	  long	handle;
	  
	  NetworkOutputStream(long handle)
	  {
		  this.handle = handle;
	  }
	  
	  @Override
	  int	captureFrom(PacketCapture in, int count) throws CapturePacketException, IOException, IllegalArgumentException
	  {
		  if (in.inputFile == null)
			  throw new IllegalArgumentException("can't capture from and to the network!");
		  // Thunk to captureToNetwork()
		  return in.captureToNetwork(count, handle);
	  }
	  
	  @Override
	  public long	getFilePointer() throws IOException
	  {
		  if (isClosed())
			  throw new IOException("capture device not open");
		  throw new IOException("capture device is not a savefile");
	  }
	  
	  @Override
	  public void	writePacket(Packet packet) throws IOException
	  {
		  if (isClosed())
			  throw new IOException("capture device not open");
		  if (packet == null)
			  throw new NullPointerException("packet is null");
		  packet.setChecksums();
		  EthernetPacket	ethernetPacket = (EthernetPacket)packet;
		  byte[]	header = ethernetPacket.getEthernetHeader();
		  byte[]	data = ethernetPacket.getEthernetData();
		  ByteBuffer	bb = ByteBuffer.allocateDirect(header.length+data.length);
		 // ByteBuffer	bb = ByteBuffer.allocate(header.length+data.length);
		  bb.put(header);
		  bb.put(data);
//		  System.out.println(this.getClass().getName()+".writePacket(): writing "+baos.size()+" bytes:"+new BigInteger(baos.toByteArray()).toString(16));
		  send0(handle, bb);
	  }
  }

  /**
   * Native support for NetworkOutputStream.
   * @param handle Handle to C++ object
   * @param bb ByteBuffer, may be direct or non-direct.
   * @since 2.0.0
   */
  private native int send0(long handle, ByteBuffer bb) throws IOException;

  /**
   * TCPDumpOutputStream class.
   * @since 2.0.0
   */
  class TCPDumpOutputStream extends PacketOutputStream implements PacketListener
  {
	  private boolean	isClosed = true;
	  File file;
	  
	  TCPDumpOutputStream(int linkType, int snaplen, File file) throws IOException
	  {
		  
		  if (this.file != null) // Close the currently open file
			  this.close();
		  
		  createDead0(handle, linkType, snaplen);
		  
		  this.file = file;
		  		  
		  openDump0(handle, file.getPath());
		  isClosed = false;
	  }
	  
	  @Override
		  int	captureFrom(PacketCapture in, int count) throws CapturePacketException, IOException, IllegalArgumentException
	  {
		  // Thunk to captureToFile()
		  return in.captureToFile(count);
	  }
	  
	  @Override
	  public void	close() throws IOException
	  {
		  if (!isClosed)
		  {
			  flush();
//			  System.out.println("TCPDumpOutputStream.close()");
			  closeDump0(handle);
		  }
		  isClosed = true;
		  saveOutputStream = null;
	  }
	  
	  @Override
	  public void	flush() throws IOException
	  {
		  if (isClosed)
			  throw new IOException("dump output stream is closed");
//		  System.out.println("TCPDumpOutputStream.flush()");
		  flushDump0(handle);
	  }
	  
	  @Override
	  public long	getFilePointer() throws IOException
	  {
		  if (isClosed())
			  throw new IOException("capture device not open");
		  return ftellDump0(handle);
	  }
	  
	  @Override
	  public void	writePacket(Packet packet) throws IOException
	  {
		  if (isClosed)
			  throw new IOException("dump output stream closed");
		  if (packet == null)
			  throw new NullPointerException("packet is null");
		  packet.setChecksums();
		  EthernetPacket	ethernetPacket = (EthernetPacket)packet;
		  byte[]	header = ethernetPacket.getEthernetHeader();
		  byte[]	data = ethernetPacket.getEthernetData();
		//  ByteBuffer	bb = ByteBuffer.allocate(header.length+data.length);
		  ByteBuffer	bb = ByteBuffer.allocateDirect(header.length+data.length);
		  bb.put(header);
		  bb.put(data);
//		  System.out.println(this.getClass().getName()+".writePacket(): writing "+baos.size()+" bytes:"+new BigInteger(baos.toByteArray()).toString(16));
		  Timeval tv = packet.getTimeval();
		  writePacket1(handle, bb,  (tv.getSeconds()*1000000L+(tv.getMicroSeconds())));
	  }

	  // PacketListener method
	  public void	packetArrived(Packet packet)
	  {
		  try
		  {
			  writePacket(packet);
		  }
		  catch (IOException exc)
		  {
			  exc.printStackTrace();
		  }
	  }
  }

  /**
   * Native support for TCPDumpOutputStream
   * @since 2.0.0
   */
  private native void	closeDump0(long handle) throws IOException;
  /**
   * Native support for TCPDumpOutputStream
   * @since 2.0.0
   */
  private native void	createDead0(long handle, int linkType, int snaplen) throws IOException;  
  /**
   * Native support for TCPDumpOutputStream
   * @since 2.0.0
   */
  private native void	flushDump0(long handle) throws IOException;
  /**
   * Native support for TCPDumpOutputStream
   * @since 2.0.0
   */
  private native long	ftellDump0(long handle) throws IOException;
  /**
   * Native support for TCPDumpOutputStream
   * @since 2.0.0
   */
  private native void	openDump0(long handle, String filename) throws IOException;
  /**
   * Native support for TCPDumpOutputStream
   * @since 2.0.0
   */
  private native void	writePacket0(long handle, ByteBuffer bb) throws IOException;
  
  private native void	writePacket1(long handle, ByteBuffer bb, long tv) throws IOException;
  
  /**
   * CompiledFilter implementation class
   *
   * @since 2.0.0
   */
  class CompiledFilterImpl extends CompiledFilter
  {
	  private String	expression;
	  private boolean	optimized;
	  long	handle;	// handle to compiled filter
	  
	  CompiledFilterImpl(String expression, boolean optimized, int netmask) throws InvalidFilterException
	  {
		  this.expression = expression;
		  this.optimized = optimized;
		  this.handle = compileFilter0(PacketCapture.this.handle, optimized, netmask, expression);
//		  System.out.println("CompiledFilterImpl.<init>() handle="+Integer.toHexString(handle));
	  }
	  
	  protected void	finalize() throws Throwable
	  {
		  try
		  {
//			  System.out.println("CompiledFilterImpl.finalize() handle="+Integer.toHexString(handle));
			  if (handle != 0)
				  freeFilter0(handle);
		  }
		  finally
		  {
			  super.finalize();
		  }
	  }
	  
	  public String	getExpression()
	  {
		  return expression;
	  }
	  
	  public boolean	isOptimized()
	  {
		  return optimized;
	  }
	  
	  long	getHandle()	{ return handle;	}
	  
//	  public String	toString()
//	  {
//		  return "CompiledFilterImpl[handle="+Integer.toHexString(handle)+"]";
//	  }
  }
  
  /**
   * Native support for CompiledFilter
   * @since 2.0.0
   */
  private native long	compileFilter0(long handle, boolean optimize, int netmask, String filter)
  throws InvalidFilterException;
  
  private native void	freeFilter0(long handle);

  // static initialization

  static {
//    System.err.print("PacketCapture: loading native library jpcap.. ");
    System.loadLibrary(LIB_PCAP_WRAPPER);
//    System.err.println("ok");
  }

  private static final String _rcsid = 
    "$Id: PacketCapture.java,v 1.7 2008/11/18 14:02:39 rickwall Exp $";
}
