// $Id: PacketDispatcher.java,v 1.2 2008/08/18 05:34:36 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Copyright (C) 2008, Esmond Pitt                                         *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.capture;

import java.util.Iterator;
import java.util.HashSet;
import java.util.Set;
import net.sourceforge.jpcap.net.Packet;
import net.sourceforge.jpcap.net.RawPacket;


/**
 * This class encapsulates a mechanism for dispatching network data and 
 * packets to a listener who has subscribed to such events.
 * <p>
 * Since this class doesn't contain an implementation for handling 
 * real network data, it is abstract.
 * <p>
 * This class is extended by packet capture systems which, although 
 * they may utilize different sources of packet data, share a common
 * mechanism for dispatching the packets.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.2 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:36 $
 */
public abstract class PacketDispatcher extends PacketHandler implements PacketDispatchCapable
{
	public void addRawPacketListener(RawPacketListener rawListener)
	{
		synchronized (rawListeners)
		{
			if (!rawListeners.add(rawListener))
				throw new IllegalArgumentException("listener already registered");
		}
	}
	
	public void removeRawPacketListener(RawPacketListener rawListener)
	{
		synchronized (rawListeners)
		{
			if (!rawListeners.remove(rawListener))
				throw new IllegalArgumentException("listener not registered");
		}
	}
	
	public void addPacketListener(PacketListener objListener)
	{
		synchronized (objListeners)
		{
			if (!objListeners.add(objListener))
				throw new IllegalArgumentException("listener already registered");
		}
	}
	
	public void removePacketListener(PacketListener objListener)
	{
		synchronized (objListeners)
		{
			if (!objListeners.remove(objListener))
				throw new IllegalArgumentException("listener not registered");
		}
	}

  /**
   * Dispatch a raw packet to all registered listeners.
   * This method operates on a copy the list of listeners,
   * so that callbacks can alter the listeners concurrently.
   */
  void dispatchRawPacket(RawPacket rawPacket)
  {
	  Iterator<RawPacketListener> i;
	  synchronized (rawListeners)
	  {
		  i = new HashSet<RawPacketListener>(rawListeners).iterator();
	  }
	  while(i.hasNext())
	  {
		  RawPacketListener pl = i.next();
		  // System.err.println(NAME + ": calling registered listener " + pl);
		  pl.rawPacketArrived(rawPacket);
	  }
  }

  /**
   * Dispatch a packet to all registered listeners.
   * This method operates on a copy the list of listeners,
   * so that callbacks can alter the listeners concurrently.
   */
  void dispatchPacket(Packet packet)
  {
	  Iterator<PacketListener> i;
	  synchronized (objListeners)
	  {
		  i = new HashSet<PacketListener>(objListeners).iterator();
	  }
	  while(i.hasNext())
	  {
		  PacketListener pl = i.next();
		  // System.err.println(NAME + ": calling registered listener " + pl);
		  pl.packetArrived(packet);
	  }
  }
  
  int getObjListenerCount()
  {
	  synchronized (objListeners)
	  {
		  return objListeners.size();
	  }
  }
  
  int getRawListenerCount()
  {
	  synchronized (rawListeners)
	  {
		  return rawListeners.size();
	  }
  }

  /**
   * Registered clients listening for raw packets captured by this object.
   */
  private Set<RawPacketListener> rawListeners = new HashSet<RawPacketListener>();

  /**
   * Registered clients listening for packet objects captured by this object.
   */
  private Set<PacketListener> objListeners = new HashSet<PacketListener>();

  private static final String _rcsid = 
    "$Id: PacketDispatcher.java,v 1.2 2008/08/18 05:34:36 esmondpitt Exp $";
}
