/*
 * PacketOutputStream.java
 *
 * Created on 10 May 2008, 11:27
 *
 * Copyright (c) Esmond Pitt, 2007. All rights reserved.
 */

package net.sourceforge.jpcap.capture;

import java.io.IOException;
import java.io.OutputStream;
import net.sourceforge.jpcap.net.Packet;

/**
 * PacketOutputStream is an I/O stream used for sending Packets to the network,
 * or dumping them to a <i>tcpdump</i>-formatted file, depending on how the stream was created.
 * <p>
 * The <code>close()</code> and <code>flush()</code> methods close and flush the output file respectively.
 * For a stream created for sending, they do nothing.
 *
 * @see PacketCapture#captureTo(int, PacketOutputStream)
 * @see PacketCapture#getOutputStream()
 * @see PacketCapture#getOutputStream(File)
 *
 * @author Esmond Pitt
 * @version $Revision: 1.1 $
 * @since 2.0.0
 */
public abstract class PacketOutputStream extends OutputStream
{
	
	/** Creates a new instance of PacketOutputStream */
	protected PacketOutputStream()
	{
	}

	/**
	 * Capture from a device.
	 * This is really an instance of the Visitor pattern,
	 * used to despatch {@link PacketCapture#captureTo(PacketOutputStream)} calls via the type of the argument.
	 * @param in PacketCapture object we are capturing from.
	 *
	 * @return the number of packets captured and written.
	 * @exception CapturePacketException Packet capture problem.
	 * @exception IOException I/O problem on output stream.
	 * @exception IllegalArgumentException if this PacketCapture object is capturing <i>from</i> the network
	 * and the output stream is dumping <i>to</i> the network, as this will just break your network.
	 *
	 * @since 2.0.0
	 */
	abstract int	captureFrom(PacketCapture in, int count) throws CapturePacketException, IOException, IllegalArgumentException;
	
	/**
	 * Return the current file position for the savefile, i.e. the number of bytes written by {@link #writePacket}.
	 * @return the current file position.
	 * @exception IOException Savefile not open, output not a savefile, etc.
	 */
	public abstract long	getFilePointer() throws IOException;
	
	/**
	 * {@inheritDoc}
	 * <p>
	 * This implementation throws {@link java.lang.UnsupportedOperationException}.
	 */
	public void	write(int b) throws IOException
	{
		throw new UnsupportedOperationException("write(int)");
	}
	
	/**
	 * {@inheritDoc}
	 * <p>
	 * This implementation throws {@link java.lang.UnsupportedOperationException}.
	 */
	public void	write(byte[] buffer) throws IOException
	{
		throw new UnsupportedOperationException("write(byte[])");
	}
	
	/**
	 * {@inheritDoc}
	 * <p>
	 * This implementation throws {@link java.lang.UnsupportedOperationException}.
	 */
	public void	write(byte[] buffer, int offset, int count) throws IOException
	{
		throw new UnsupportedOperationException("write(byte[], int, int)");
	}
	
	/**
	 * Write a packet, by either:
	 * <ul>
	 * <li>sending it to the network, or
	 * <li>dumping it to a <i>tcpdump</i>-formatted output file,
	 * </ul>depending on how the stream was created.
	 * The packet is written after calling {@link Packet#setChecksums()},
	 * so checksums &amp;c are <i>not</i> the application's problem,
	 * at least for packet classes which correctly
	 * implement {@link Packet#setChecksums()}.
	 *
	 * @param packet Packet to write.
	 * @exception IOException Some I/O problem sending or dumping the packet.
	 */
	public abstract void	writePacket(Packet packet) throws IOException;
	
}
