// $Id: ARPPacket.java,v 1.2 2008/08/18 05:34:37 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.net;

import net.sourceforge.jpcap.util.AnsiEscapeSequences;
import net.sourceforge.jpcap.util.Timeval;
import java.net.InetAddress;
import java.nio.ByteBuffer;

/**
 * An ARP protocol packet.
 * <p>
 * Extends EthernetPacket, adding ARP header information and an ARP 
 * data payload.
 *<p>Now handles both ARP and RARP. 
 *
 * @author Patrick Charles
 * @author Jonas Lehmann
 * @author Esmond Pitt
 * @version $Revision: 1.2 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:37 $
 */
public class ARPPacket extends EthernetPacket implements ARPFields
{
	static final long serialVersionUID = 2311191227616232702L;

	/*
	static
	{
		assert ARP_HW_TYPE_POS == 0 : "ARPField ARP_HW_TYPE_POS="+ARP_HW_TYPE_POS;
		assert ARP_PR_TYPE_POS == 2 : "ARPField ARP_PR_TYPE_POS="+ARP_PR_TYPE_POS;
		assert ARP_HW_LEN_POS == 4 : "ARPField ARP_HW_LEN_POS="+ARP_HW_LEN_POS;
		assert ARP_PR_LEN_POS == 5 : "ARPField ARP_PR_LEN_POS="+ARP_PR_LEN_POS;
		assert ARP_OP_POS == 6 : "ARPField ARP_OP_POS="+ARP_OP_POS;
		assert ARP_S_HW_ADDR_POS == 8 : "ARPField ARP_S_HW_ADDR_POS="+ARP_S_HW_ADDR_POS;
		assert ARP_S_PR_ADDR_POS == 14 : "ARPField ARP_S_PR_ADDR_POS="+ARP_S_PR_ADDR_POS;
		assert ARP_T_HW_ADDR_POS == 18 : "ARPField ARP_T_HW_ADDR_POS="+ARP_T_HW_ADDR_POS;
		assert ARP_T_PR_ADDR_POS == 24 : "ARPField ARP_T_PR_ADDR_POS="+ARP_T_PR_ADDR_POS;
		assert ARP_HEADER_LEN == 28 : "ARPField offsets and lengths don't add up to 28";
	}
	 */

	// EJP 7 May 2008: revised constructors not to auto-extract header and data, to economize on memory.
	// EJP 29 May 2008: revised to use ByteBuffer throughout.
	
  /**
   * Create a new ARP packet.
   * @param lLen link-layer header length
   * @param bytes Packet data
   */
  public ARPPacket(int lLen, byte [] bytes) {
    this(lLen, bytes, null);
  }

  /**
   * Create a new ARP packet.
   * @param lLen link-layer header length
   * @param bytes Packet data
   * @param tv Timestamp
   */
  public ARPPacket(int lLen, byte [] bytes, Timeval tv) {
	this(lLen, ByteBuffer.wrap(bytes), tv);
  }

  /**
   * Create a new ARP packet.
   * @param lLen link-layer header length
   * @param bb Packet data wrapped in a ByteBuffer
   * @param tv Timestamp
   * @since 2.0.0
   */
  public ARPPacket(int lLen, ByteBuffer bb, Timeval tv) {
    super(lLen, bb, tv);
	// EJP 7 May 2008: revised this not to auto-extract header and data, to economize on memory.
  }
  
  /**
   * Return the hardware address type in this packet,
   * usually {@link ARPFields#ARP_ETH_ADDR_CODE}.
   * @return the hardware address type in this packet.
   * @since 2.0.0
   */
  public int	getHwAddressType()
  {
	  return _byteBuffer.getShort(_ethOffset+ARP_HW_TYPE_POS) & 0xffff;
  }
  
  /**
   * Return the length of the hardware addresses in this packet,
   * usually 6 for {@link ARPFields#ARP_ETH_ADDR_CODE}.
   * @return the length of the hardware addresses in this packet.
   * @since 2.0.0
   */
  public int	getHwAddressLength()
  {
	  return _byteBuffer.get(_ethOffset+ARP_HW_LEN_POS);
  }
  
  /**
   * Return the type of the protocol addresses in this packet,
   * usually {@link ARPFields#ARP_IP_ADDR_CODE}.
   * @return the type of the protocol addresses in this packet.
   * @since 2.0.0
   */
  public int	getProtoAddressType()
  {
	  return _byteBuffer.getShort(_ethOffset+ARP_PR_TYPE_POS) & 0xffff;
  }
  
  /**
   * Return the length of the protocol addresses in this packet,
   * usually 4 for {@link ARPFields#ARP_IP_ADDR_CODE}.
   * @return the length of the protocol addresses in this packet.
   * @since 2.0.0
   */
  public int	getProtoAddressLength()
  {
	  return _byteBuffer.get(_ethOffset+ARP_PR_LEN_POS);
  }

  /**
   * Fetch the ARP/RARP operation code.
   * Usually one of ARPFields.{ARP_OP_REQ_CODE, ARP_OP_REP_CODE, RARP_REQ_CODE, RARP_REP_CODE}.
   */
  public int getOperation() {
//    return ArrayHelper.extractInteger(getARPHeader(), ARP_OP_POS, ARP_OP_LEN);
	  return _byteBuffer.getShort(_ethOffset+ARP_OP_POS) & 0xffff;
  }
  
  /**
   * Set the ARP operation code.
   * Usually one of ARPFields.{ARP_OP_REQ_CODE, ARP_OP_REP_CODE, RARP_REQ_CODE, RARP_REP_CODE}.
   * @param operation New operation code: not checked for validity.
   * @since 2.0.0
   */
  public void	setOperation(int operation)
  {
	  _byteBuffer.putShort(_ethOffset+ARP_OP_POS, (short)operation);
  }

  /**
   * {@inheritDoc}
   *<p>
   * This method also calls {@link EthernetPacket#setSourceMACAddress super.setSourceMACAddress()}
   * to set the corresponding field in the Ethernet header.
   *
   * @since 2.0.0
   */
  @Override
  public void	setSourceMACAddress(MACAddress macAddress)
  {
	  _byteBuffer.position(_ethOffset+ARP_S_HW_ADDR_POS);
	  _byteBuffer.put(macAddress.getAddress());
	  super.setSourceMACAddress(macAddress);
  }

  /**
   * Fetch the sender protocol address.
   */
  public String getSourceProtoAddress() {
//    return IPAddress.extract(ARP_S_PR_ADDR_POS, getARPHeader());
	  byte[]	bytes = new byte[IPAddress.WIDTH];
	  _byteBuffer.position(_ethOffset+ARP_S_PR_ADDR_POS);
	  _byteBuffer.get(bytes);
	  return IPAddress.extract(0, bytes);
  }
  
  /**
   * Set the source protocol address.
   * @param address Source IP address.
   * @since 2.0.0
   */
  public void	setSourceProtoAddress(InetAddress address)
  {
	  if (address.getAddress().length != getProtoAddressLength())
		  throw new IllegalArgumentException("Invalid protocol address length: expected "+getProtoAddressLength()+" got "+address.getAddress().length);
	  _byteBuffer.position(_ethOffset+ARP_S_PR_ADDR_POS);
	  _byteBuffer.put(address.getAddress());
  }
  
  /**
   * {@inheritDoc}
   *<p>
   * This method also calls {@link EthernetPacket#setDestinationMACAddress super.setDestinationMACAddress()}
   * to set the corresponding field in the Ethernet header.
   *
   * @since 2.0.0
   */
  @Override
  public void	setDestinationMACAddress(MACAddress macAddress)
  {
	  _byteBuffer.position(_ethOffset+ARP_T_HW_ADDR_POS);
	  _byteBuffer.put(macAddress.getAddress());
	  super.setDestinationMACAddress(macAddress);
  }

  /**
   * Fetch the destination protocol address.
   */
  public String getDestinationProtoAddress() {
//    return IPAddress.extract(ARP_T_PR_ADDR_POS, getARPHeader());
	  byte[]	bytes = new byte[IPAddress.WIDTH];
	  _byteBuffer.position(_ethOffset+ARP_T_PR_ADDR_POS);
	  _byteBuffer.get(bytes);
	  return IPAddress.extract(0, bytes);
  }
  
  /**
   * Set the destination protocol address.
   * @param address Destination IP address.
   * @since 2.0.0
   */
  public void	setDestinationProtoAddress(InetAddress address)
  {
	  if (address.getAddress().length != getProtoAddressLength())
		  throw new IllegalArgumentException("Invalid protocol address length: expected "+getProtoAddressLength()+" got "+address.getAddress().length);
	  _byteBuffer.position(_ethOffset+ARP_T_PR_ADDR_POS);
	  _byteBuffer.put(address.getAddress());
  }

  /**
   * Fetch the arp header, excluding arp data payload.
   * @return the ARP header
   */
  public byte [] getARPHeader() {
//    return PacketEncoding.extractHeader(_ethOffset, ARP_HEADER_LEN, _bytes);
	  byte[]	bytes = new byte[ARP_HEADER_LEN];
	  _byteBuffer.position(_ethOffset);
	  _byteBuffer.get(bytes);
	  return bytes;
  }

  /**
   * Fetch data portion of the arp packet.
   * @return the ARP data
   */
  public byte [] getARPData() {
//    return PacketEncoding.extractData(_ethOffset, ARP_HEADER_LEN, _bytes);
	  _byteBuffer.position(_ethOffset+ARP_HEADER_LEN);
	  byte[]	bytes = new byte[_byteBuffer.remaining()];
	  _byteBuffer.get(bytes);
	  return bytes;
  }

  /**
   * {@inheritDoc}
   */
  public byte [] getHeader() {
    return getARPHeader();
  }

  /**
   * {@inheritDoc}
   */
  public byte [] getData() {
    return getARPData();
  }

  /**
   * {@inheritDoc}
   */
  public String toColoredString(boolean colored) {
    StringBuffer buffer = new StringBuffer();
    buffer.append('[');
    if(colored) buffer.append(getColor());
    buffer.append("ARPPacket");
    if(colored) buffer.append(AnsiEscapeSequences.RESET);
    
	buffer.append(": ");
	
	switch (super.getEthernetProtocol())
	{
		case EthernetProtocols.ARP:
			buffer.append("ARP ");
		    buffer.append(getOperation() == ARP_OP_REQ_CODE ? "request" : "reply");
			break;
		case EthernetProtocols.RARP:
			buffer.append("RARP ");
		    buffer.append(getOperation() == RARP_OP_REQ_CODE ? "request" : "reply");
			break;
		default:
			buffer.append("Unknown ARP/RAP protocol 0x"+Integer.toHexString(getEthernetProtocol())+" operation="+getOperation());
			break;
	}
    buffer.append(' ');
    buffer.append(getSourceMACAddress().toString() + "->" + 
                  getDestinationMACAddress().toString());
    buffer.append(", ");
    buffer.append(getSourceProtoAddress() + "->" + 
                  getDestinationProtoAddress());
    buffer.append(" l=" + getARPHeader().length + "," + getARPData().length);
	
	/*
	buffer.append(":header={0x"+new BigInteger(1, getARPHeader()).toString(16)+"}");
	buffer.append(":data={0x"+new BigInteger(1, getARPData()).toString(16)+"}");
	buffer.append(":packet={0x"+new BigInteger(1, _bytes).toString(16)+"}");
	*/
	
    buffer.append(']');

    return buffer.toString();
  }

  /**
   * {@inheritDoc}
   */
  public String getColor() {
    return AnsiEscapeSequences.PURPLE;
  }

  private static final String _rcsid = 
    "$Id: ARPPacket.java,v 1.2 2008/08/18 05:34:37 esmondpitt Exp $";
}
