/*
 * IP.java
 *
 * Created on 30 May 2008, 12:35
 *
 * Copyright (c) Esmond Pitt, 2007. All rights reserved.
 */

package net.sourceforge.jpcap.net;

import java.net.InetAddress;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import net.sourceforge.jpcap.util.AnsiEscapeSequences;
import net.sourceforge.jpcap.util.Timeval;

/**
 * IP interface.
 *
 * @author Esmond Pitt
 * @version $Revision: 1.1 $
 * @since 2.0.0
 */
public interface IP
{
	/**
	 * Base IP packet class.
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 *
	 * @since 2.0.0.
	 */
	abstract class Packet extends EthernetPacket
	{
		static final long serialVersionUID = 5054015884523628964L;
	
		/**
		 * offset from beginning of byte array where IP header ends (i.e.,
		 * size of ethernet frame header and IP header).
		 */
		protected int _ipOffset;
		/**
		 * Length of IP header).
		 */
		protected int _ipHeaderLength;
		
		protected Packet(int lLen, byte[] bytes)
		{
			this(lLen, bytes, null);
		}
		
		protected Packet(int lLen, byte[] bytes, Timeval tv)
		{
			this(lLen, ByteBuffer.wrap(bytes), tv);
		}
		
		protected Packet(int lLen, ByteBuffer bb, Timeval tv)
		{
			super(lLen, bb, tv);
		    this._ipOffset = _ethOffset + getIPHeaderLength();
			// set offset into _bytes of previous layers
			this._ipHeaderLength = getIPHeaderLength();
		}
		
		/**
		 * Fetch the IP header length in bytes.
		 * @return the IP header length in bytes
		 */
		public abstract int getIPHeaderLength();
		/**
		 * Get the IP version code.
		 * @return the IP version (4 or 6).
		 */
		public abstract int getVersion();
		/**
		 * Fetch the code indicating the type of protocol embedded in the IP packet.
		 * @return the IP protocol identifier.
		 * @see IPProtocols
		 */
		public abstract int getIPProtocol();
		/**
		 * Fetch the source IP address.
		 * @return the source IP address
		 * @since 2.0.0
		 */
		public abstract InetAddress getSourceInetAddress();
		/**
		 * Set the source address.
		 * @param address Source address.
		 * @since 2.0.0
		 */
		public abstract void	setSourceInetAddress(InetAddress address);
		/**
		 * Fetch the destination IP address.
		 * @return the destination IP address
		 * @since 2.0.0
		 */
		public abstract InetAddress getDestinationInetAddress(); 
		/**
		 * Set the destination address.
		 * @param address Destination address.
		 * @since 2.0.0
		 */
		public abstract void	setDestinationInetAddress(InetAddress address);
		/**
		 * Fetch the IP header as a byte array.
		 * @return the IP header
		 */
		public abstract byte[] getIPHeader();
		/**
		 * Fetch the IP data as a byte array.
		 * @return the IP payload
		 */
		public abstract byte[] getIPData();

		/**
		 * {@inheritDoc}
		 */
		public int getHeaderLength()
		{
			return getIPHeaderLength();
		}
		
		/**
		 * Fetch the IP payload length in bytes.
		 * @return the IP payload length in bytes.
		 */
		public abstract int getLength();
		
		/**
		 * {@inheritDoc}
		 */
		public byte[] getHeader()
		{
			return getIPHeader();
		}
		
		/**
		 * {@inheritDoc}
		 */
		public byte[] getData()
		{
			return getIPData();
		}

		/**
		 * {@inheritDoc}
		 */
		public String toColoredString(boolean colored)
		{
			StringBuffer buffer = new StringBuffer();
			buffer.append('[');
			if(colored) buffer.append(getColor());
			buffer.append("IPv"+getVersion()+"Packet");
			if(colored) buffer.append(AnsiEscapeSequences.RESET);
			buffer.append(": ");
			buffer.append(getSourceInetAddress().getHostAddress() + "->" +
				getDestinationInetAddress().getHostAddress());
			buffer.append(" proto=" + getProtocol());
			buffer.append(" l=" + getIPHeaderLength() + "," + getLength());
			buffer.append(']');
			
			return buffer.toString();
		}
		/**
		 * Convert this IP packet to a more verbose string.
		 */
		public abstract String toColoredVerboseString(boolean colored);
		
		/**
		 * {@inheritDoc}
		 */
		public String getColor()
		{
			return AnsiEscapeSequences.WHITE;
		}
	}
	
	/**
	 * ICMP (Internet Control Message Protocol) API interface.
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 *
	 * @since 2.0.0
	 */
	interface ICMP
	{
		/**
		 * Fetch the ICMP header as a byte array.
		 * Modifying this byte array has no effect on the packet.
		 * @return the ICMP header.
		 */
		byte[] getICMPHeader();
		/**
		 * Fetch the ICMP data as a byte array.
		 * Modifying this byte array has no effect on the packet.
		 * @return the ICMP data
		 */
		byte[] getICMPData();
		/**
		 * Fetch the ICMP message type, including subcode. Return value can be
		 * used with ICMPMessage.getDescription().
		 * @return a 2-byte value containing the message type in the high byte
		 * and the message type subcode in the low byte.
		 */
		int getMessageCode();
		/**
		 * Set the ICMP message code and subcode (16 bits) as per {@link #getMessageCode}.
		 * @param code 2-byte code value.
		 */
		void	setMessageCode(short code);
		/**
		 * Fetch the ICMP message type code.
		 * @return the ICMP message type code.
		 * @since 2.0.0.
		 */
		int getMessageMajorCode();
		/**
		 * Fetch the ICMP message subcode.
		 * @return the ICMP message subcode.
		 */
		int getMessageMinorCode();
		/**
		 * Fetch the ICMP checksum.
		 * @return the ICMP checksum.
		 */
		int getICMPChecksum();
	}
	
	/**
	 * TCP (Transmission Control Protocol) API interface.
	 * This is read-only. You can't use Jpcap to send TCP packets, because
	 *<ol>
	 *<li> the TCP at the other end will ignore them unless you get the TCP sequence numbers correct
	 *<li> the local TCP will reject the replies with an RST unless the sequence numbers, port, etc
	 * all correspond to a real TCP application at this end: in which case you don't need Jpcap anyway.
	 *</ol>
	 * ... and even if you could, what exactly is the point, when you have a nice TCP API already?
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 *
	 * @since 2.0.0
	 */
	interface TCP
	{
		/**
		 * Fetch the source port number.
		 * @return the souce port number.
		 */
		int getSourcePort();
		/**
		 * Fetches the destination port number.
		 * @return the destination port number.
		 */
		int getDestinationPort();
		/**
		 * Fetch the byte sequence number.
		 * @return the byte sequence number.
		 */
		long getSequenceNumber();
		/**
		 * Fetch the byte acknowledgment number, i.e. the next byte that the sender expects to receive.
		 * @return the byte acknowledgement number.
		 */
		long getAcknowledgmentNumber();
		/**
		 * Fetch the TCP header length in bytes.
		 */
		int getTCPHeaderLength();
		/**
		 * Fetches the length of the payload data.
		 * @return the length of the payload data.
		 */
		int getPayloadDataLength();
		/**
		 * Fetch the window size.
		 * @return the window size.
		 */
		int getWindowSize();
		/**
		 * Fetch the TCP checksum.
		 * @return the TCP checksum.
		 */
		int getTCPChecksum();
		/**
		 * Fetch the urgent pointer.
		 * @return the TCP urgent pointer
		 */
		int getUrgentPointer();
		/**
		 * Return the URG flag, indicating that the urgent pointer is valid.
		 * @return the URG flag.
		 */
		boolean isUrg();
		/**
		 * Return the ACK flag, indicating that the ack number is valid.
		 * @return the ACK flag.
		 */
		boolean isAck();
		/**
		 * Return the PSH flag, indicating that the receiver should pass the
		 * data to the application as soon as possible.
		 * @return the PSH flag.
		 */
		boolean isPsh();
		/**
		 * Return the RST flag, indicating that the connection has been reset by the sender.
		 * @return the RST flag.
		 */
		boolean isRst();
		/**
		 * Return the SYN flag, set to synchronize the sequence numbers
		 * between the sender and receiver as part of the connect handshake.
		 * @return the SYN flag.
		 */
		boolean isSyn();
		/**
		 * Return the FIN flag, indicating that the sender has finished sending.
		 * @return the FIN flag.
		 */
		boolean isFin();
		/**
		 * Fetch the TCP header as a byte array.
		 * Altering this has no effect on the packet.
		 */
		byte[] getTCPHeader();
		/**
		 * Fetch the TCP data as a byte array.
		 * Altering this has no effect on the packet.
		 */
		byte[] getTCPData();
		/**
		 * Return true if the TCP checksum is valid
		 * @return true or false.
		 */
		boolean	isValidTCPChecksum();
		
	}
	
	/**
	 * UDP (User Datagram Protocol) API interface.
	 * This is read-only. You probably can use Jpcap to send UDP packets,
	 * but what exactly is the point, when you have a nice TCP API already?
	 *
	 * @author Esmond Pitt
	 * @version $Revision: 1.1 $
	 *
	 * @since 2.0.0
	 *
	 */
	interface UDP
	{
		/**
		 * Fetch the source port number.
		 * @return the souce port number.
		 */
		int getSourcePort();
		/**
		 * Fetches the destination port number.
		 * @return the destination port number.
		 */
		int getDestinationPort();
		/**
		 * Fetch the UDP checksum.
		 * @return the UDP checksum.
		 */
		int getUDPChecksum();
		/**
		 * Fetch the UDP header as a byte array.
		 * @return the UDP header as a byte array.
		 */
		byte[] getUDPHeader();
		/**
		 * Fetch the UDP data as a byte array.
		 * @return the UDP data as a byte array.
		 */
		byte[] getUDPData();
		/**
		 * Return true iff the UDP checksum is valid.
		 * @return true or false.
		 */
		boolean	isValidUDPChecksum();
	}
}
