// $Id: MACAddress.java,v 1.2 2008/08/18 05:34:37 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.net;
import java.math.BigInteger;
import net.sourceforge.jpcap.util.HexHelper;


/**
 * MAC address.
 * <p>
 * This class encapsulates MAC addresses, and provides utility methods
 * to extract a MAC address from a big-endian byte array or String.
 *
 * @author Patrick Charles and Jonas Lehmann
 * @author EJP
 * @version $Revision: 1.2 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:37 $
 */
public class MACAddress
{
	private byte[]	address;

	/**
	 * Construct a new instance of MACAddress.
	 * @param address Address.
	 * @since 2.0.0
	 */
	public MACAddress(byte[] address)
	{
		if (address.length != WIDTH)
			throw new IllegalArgumentException("Incorrect MAC address size");
		this.address = address;
	}
	
	/**
	 * Construct a new instance of MACAddress.
	 * @param address formatted Address, as per {@link #getMACAddress()}.
	 * @since 2.0.0
	 */
	public MACAddress(String address)
	{
		this.address = extract(address);
	}
	
	/**
	 * Return the MAC address as a byte[] array.
	 * To get the MAC address as a string, call {@link #toString}.
	 * @return a byte array.
	 * @see #getAddressString()
	 * @see #getMACAddress()
	 * @see #toString()
	 * @since 2.0.0
	 */
	public	byte[]	getAddress()
	{
		return address;
	}
	
	/**
	 * Fetch the MAC address as an unformatted string.
	 * @return an unformatted string, to match {@link java.net.NetworkInterface}.
	 *
	 * @since 2.0.0
	 */
	public String	getAddressString()
	{
		// Ensure leading zeroes
		String	s = "000000000000"+new BigInteger(1, address).toString(16);
		return s.substring(s.length()-12);
	}
	
	/**
	 * Fetch the MAC address as formatted string.
	 * @return a formatted MAC address string, e.g. 00:11:22:33:44:55.
	 *
	 * @since 2.0.0
	 */
	public String	getMACAddress()
	{
		return extract(0, address);
	}
	
	/**
	 * {@inheritDoc}
	 *<p>
	 * This method returns the result of {@link #getMACAddress()}.
	 * @since 2.0.0
	 */
	public String	toString()
	{
		return getMACAddress();
	}
	
  /**
   * Extract a MAC address from an array of bytes.
   * @param offset the offset of the address data from the start of the 
   * packet.
   * @param bytes an array of bytes containing at least one MAC address.
   */
	public static String extract(int offset, byte [] bytes)
	{
		StringBuffer sa = new StringBuffer();
		for(int i = offset; i < offset+WIDTH; i++)
		{
			if (i > offset)
				sa.append(SEPARATOR);
			sa.append(HexHelper.toString(bytes[i]));
		}
		return sa.toString();
	}
	
  /**
   * Extract a MAC address from a String.
   * @param address a MAC-address string as returned by {@link #extract(int, byte[])}
   * @return a byte array
   * @since 2.0.0
   */
  public static byte[] extract(String address)
  {
	  if (address.length() != WIDTH*3-1)
		  throw new IllegalArgumentException("Incorrect MAC address size");
	  byte[]	bytes = new byte[WIDTH];
	  for (int i = 0; i < WIDTH; i++)
	  {
		  if (i > 0)
			  if (address.charAt(i*3-1) != SEPARATOR)
				  throw new IllegalArgumentException("Incorrect MAC address format");
		  bytes[i] = (byte)Integer.parseInt(address.substring(i*3, i*3+2), 16);
	  }
	  return bytes;
  }

  /**
   * Generate a random MAC address.
   */
  public static long random() {
    return (long)(0xffffffffffffL * Math.random());
  }

  /**
   * The width in bytes of a MAC address.
   */
  public static final int WIDTH = 6;
  /**
   * The separator character.
   */
  public static final char	SEPARATOR = ':';

  private static final String _rcsid = 
    "$Id: MACAddress.java,v 1.2 2008/08/18 05:34:37 esmondpitt Exp $";

  /*
  public static void	main(String[] args)
  {
	  MACAddress	m1 = new MACAddress("00:11:22:33:44:55");
	  System.out.println(m1);
	  MACAddress	m2 = new MACAddress(new byte[]{0x00,0x11,0x22,0x33,0x44,0x55});
	  System.out.println(m2);
	  assert Arrays.equals(m1.getAddress(), m2.getAddress()) : "MACAddress test failed";
  }
   */
}
