// $Id: TCPPacket.java,v 1.3 2008/10/29 18:55:16 rickwall Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.net;

import net.sourceforge.jpcap.util.AnsiEscapeSequences;
import net.sourceforge.jpcap.util.Convert;
import net.sourceforge.jpcap.util.Timeval;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;


/**
 * A TCP packet.
 * <p>
 * Extends an IP packet, adding a TCP header and TCP data payload.
 * Please use IPv4.TCP.Packet for new code.
 *
 * @param IP IP packet implementation
 *
 * @author Patrick Charles
 * @author Jonas Lehmann
 * @author Esmond Pitt
 *
 * @version $Revision: 1.3 $
 * @lastModifiedBy $Author: rickwall $
 * @lastModifiedAt $Date: 2008/10/29 18:55:16 $
 */
public class TCPPacket extends IPPacket implements TCPFields, IP.TCP
{
	static final long serialVersionUID = 756511080662883271L;

	private int _tcpHeaderLength;
  
  /** 
   * Create a new TCP packet.
   * @param lLen link-layer header length
   * @param bytes Packet data
   */
  public TCPPacket(int lLen, byte [] bytes) {
    this(lLen,bytes, null);
  }

  /**
   * Create a new TCP packet.
   * @param lLen link-layer header length
   * @param bytes Packet data
   * @param tv Timestamp
   */
  public TCPPacket(int lLen, byte [] bytes, Timeval tv) {
    this(lLen, ByteBuffer.wrap(bytes), tv);
  }

  public TCPPacket(int lLen, ByteBuffer bb, Timeval tv) {
    super(lLen, bb, tv);
	// set TCP header length
	try
	{
		_tcpHeaderLength = ((_byteBuffer.getShort(_ipOffset + TCP_FLAG_POS) >> 12) & 0xf) * 4;
	}
	catch (IndexOutOfBoundsException exc)
	{
		
	}
  }

  /** 
   * Fetch the byte acknowledgment number. 
   * @deprecated Please use {@link #getAcknowledgmentNumber}
   */
  @Deprecated
  public long getAcknowledgementNumber() {
    return getAcknowledgmentNumber();
  }

  /** 
   * Fetch the TCP header length in bytes.
   * @deprecated Please use {@link #getTCPHeaderLength}
   */
  @Deprecated
  public int getTcpHeaderLength() {
    // this is the old method call, but everything else uses all caps for
    // TCP, so in the interest of consistency...
    return getTCPHeaderLength();
  }
			
  public int getSourcePort()
  {
	  return _byteBuffer.getShort(_ipOffset + TCP_SP_POS) & 0xffff;
  }
  
  public int getDestinationPort()
  {
	  return _byteBuffer.getShort(_ipOffset + TCP_DP_POS) & 0xffff;
  }
  
  public long getSequenceNumber()
  {
	  return _byteBuffer.getInt(_ipOffset + TCP_SEQ_POS) & 0xffffffffL;
  }
  
  public long getAcknowledgmentNumber()
  {
	  return _byteBuffer.getInt(_ipOffset + TCP_ACK_POS) & 0xffffffffL;
  }
  
  public int getTCPHeaderLength()
  {
	  return _tcpHeaderLength;
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
	  public int getHeaderLength()
  {
	  return getTCPHeaderLength();
  }
  
  public int getPayloadDataLength()
  {
	  // set data (payload) length based on info in headers (note: tcpdump
	  //  can return extra junk bytes which bubble up to here
	  int tmpLen = getLength() - getIPHeaderLength() - _tcpHeaderLength;
	  return tmpLen < 0 ? 0 : tmpLen;
  }
  
  public int getWindowSize()
  {
	  return _byteBuffer.getShort(_ipOffset+TCP_WIN_POS) & 0xffff;
  }
  
  public int getTCPChecksum()
  {
	  return _byteBuffer.getShort(_ipOffset+TCP_CSUM_POS) & 0xffff;
  }
  
  /**
   * {@inheritDoc}
   */
  public int getChecksum()
  {
	  return getTCPChecksum();
  }
  
  public int getUrgentPointer()
  {
	  return _byteBuffer.getShort(_ipOffset+TCP_URG_POS) & 0xffff;
  }
  
  private int getAllFlags()
  {
	  return _byteBuffer.getShort(_ipOffset+TCP_FLAG_POS) & 0xffff;
  }
  
  // EJP 7 May 2008: deleted ten booleans that were there to manage five bits!
  // Quicker and smaller to just test the appropriate flag bit.
  // Similarly, removed many other redundant xxxSet booleans that could be deduced
  // from an impossible value for xxx.
  
  public boolean isUrg()
  {
	  return (getAllFlags() & TCP_URG_MASK) != 0;
  }
  
  public boolean isAck()
  {
	  return (getAllFlags() & TCP_ACK_MASK) != 0;
  }
  
  public boolean isPsh()
  {
	  return (getAllFlags() & TCP_PSH_MASK) != 0;
  }
  
  public boolean isRst()
  {
	  return (getAllFlags() & TCP_RST_MASK) != 0;
  }
  
  public boolean isSyn()
  {
	  return (getAllFlags() & TCP_SYN_MASK) != 0;
  }
  
  public boolean isFin()
  {
	  return (getAllFlags() & TCP_FIN_MASK) != 0;
  }
  
  public byte[] getTCPHeader()
  {
	  byte[]	data = new byte[getTCPHeaderLength()];
	  try
	  {
		  _byteBuffer.position(_ipOffset);
		  _byteBuffer.get(data);
	  }
	  catch (IllegalArgumentException exc)
	  {
		  
	  }
	  catch (BufferUnderflowException exc)
	  {
		  
	  }
	  return data;
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
	  public byte[] getHeader()
  {
	  return getTCPHeader();
  }
  
  public byte[] getTCPData()
  {
	  // set data length based on info in headers (note: tcpdump
	  //  can return extra junk bytes which bubble up to here
	  try
	  {
		  _byteBuffer.position(_ipOffset+getTCPHeaderLength());
	  }
	  catch (IllegalArgumentException exc)
	  {
	  }
	  catch (BufferUnderflowException exc)
	  {
	  }
	  byte[]	data = new byte[_byteBuffer.remaining()];
	  _byteBuffer.get(data);
	  return data;
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
	  public byte[] getData()
  {
	  return getTCPData();
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
	  public boolean	isValidChecksum()
  {
	  return super.isValidChecksum() && isValidTCPChecksum();
  }
  
  public boolean	isValidTCPChecksum()
  {
	  return getTCPChecksum() == computeTCPChecksum();
  }

  protected int computeTCPChecksum() {
		// First thing we need to do is create the temporary buffer and rebuild
		// the header.

		// The pseudo header looks like this.
		// Source IP + Destination IP + zero + protocol + tcp Length
		// a PADDING of zero bits is added to make the segment a multiple of 16 bits

		byte[] sourceAddress = new byte[IPAddress.WIDTH];
		_byteBuffer.position(_ethOffset + IP_SRC_POS);
		_byteBuffer.get(sourceAddress);

		byte[] destAddress = new byte[IPAddress.WIDTH];
		_byteBuffer.position(_ethOffset+IP_DST_POS);
		_byteBuffer.get(destAddress);

		byte ipProto;
		ipProto = _byteBuffer.get(_ethOffset+IP_CODE_POS);

		byte[] ipData = getIPData();
		ipData[TCP_CSUM_POS] = 0x0;
		ipData[TCP_CSUM_POS+1] = 0x0;
				
		// Generate pseudo header
		byte[] pseudoHeader = new byte[ipData.length + 12];
		ByteBuffer buffer = ByteBuffer.wrap(pseudoHeader);
		buffer.putInt(Convert.bytesToInt(sourceAddress));
		buffer.putInt(Convert.bytesToInt(destAddress));
		buffer.put((byte) 0x00); // Not sure what this is.
		buffer.put((byte) ipProto);
		buffer.put((byte) ((ipData.length >> 8) & 0xff));
		buffer.put((byte) (ipData.length & 0xff));
		buffer.put(ipData,0,ipData.length);

		return ipChecksum(pseudoHeader,0, pseudoHeader.length);
	}
  
  /**
   * {@inheritDoc}
   */
  @Override
	  public void	setChecksums()
  {
	  // The checksum is over all the data from the IP SRC and DST addresses to the entire TCP header and data
	  int checksum = computeTCPChecksum();
	  _byteBuffer.putShort(_ipOffset+TCP_CSUM_POS, (short)checksum);
	  super.setChecksums();
  }
  
  /**
   * {@inheritDoc}
   */
  public String toColoredString(boolean colored)
  {
	  StringBuffer buffer = new StringBuffer();
	  buffer.append('[');
	  if(colored) buffer.append(getColor());
	  buffer.append("TCPv4Packet");
	  if(colored) buffer.append(AnsiEscapeSequences.RESET);
	  buffer.append(": ");
	  buffer.append(getSourceInetAddress().getHostAddress());
	  buffer.append('.');
	  buffer.append(IPPort.getName(getSourcePort()));
	  buffer.append("->");
	  buffer.append(getDestinationInetAddress().getHostAddress());
	  buffer.append('.');
	  buffer.append(IPPort.getName(getDestinationPort()));
	  if(isUrg())
		  buffer.append(";urg[0x" + Integer.toHexString(getUrgentPointer()) + "]");
	  if(isAck())
		  buffer.append(";ack[0x" +
			  Long.toHexString(getAcknowledgmentNumber()) + "]");
	  if(isPsh()) buffer.append(";psh");
	  if(isRst()) buffer.append(";rst");
	  if(isSyn())
		  buffer.append(";syn[0x" +
			  Long.toHexString(getSequenceNumber()) + "]");
	  if(isFin()) buffer.append(";fin");
	  buffer.append(" l=" + getTCPHeaderLength() + "," +
		  getPayloadDataLength());
	  buffer.append(']');
	  
	  return buffer.toString();
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
	  public String toColoredVerboseString(boolean colored)
  {
	  StringBuffer buffer = new StringBuffer();
	  buffer.append('[');
	  if(colored) buffer.append(getColor());
	  buffer.append("TCPPacket");
	  if(colored) buffer.append(AnsiEscapeSequences.RESET);
	  buffer.append(": ");
	  buffer.append("sport=" + getSourcePort() + ", ");
	  buffer.append("dport=" + getDestinationPort() + ", ");
	  buffer.append("seqn=0x" + Long.toHexString(getSequenceNumber()) + ", ");
	  buffer.append("ackn=0x" +
		  Long.toHexString(getAcknowledgmentNumber()) + ", ");
	  buffer.append("hlen=" + getHeaderLength() + ", ");
	  buffer.append("urg=" + isUrg() + ", ");
	  buffer.append("ack=" + isAck() + ", ");
	  buffer.append("psh=" + isPsh() + ", ");
	  buffer.append("rst=" + isRst() + ", ");
	  buffer.append("syn=" + isSyn() + ", ");
	  buffer.append("fin=" + isFin() + ", ");
	  buffer.append("wsize=" + getWindowSize() + ", ");
	  buffer.append("sum=0x" + Integer.toHexString(getChecksum()) + ", ");
	  buffer.append("uptr=0x" + Integer.toHexString(getUrgentPointer()));
	  buffer.append(']');
	  
	  return buffer.toString();
  }
  
  /**
   * {@inheritDoc}
   */
  public String getColor()
  {
	  return AnsiEscapeSequences.YELLOW;
  }

  private static final String _rcsid = 
    "$Id: TCPPacket.java,v 1.3 2008/10/29 18:55:16 rickwall Exp $";
}
