// $Id: UDPPacket.java,v 1.2 2008/08/18 05:34:38 esmondpitt Exp $

/***************************************************************************
 * Copyright (C) 2001, Patrick Charles and Jonas Lehmann                   *
 * Distributed under the Mozilla Public License                            *
 *   http://www.mozilla.org/NPL/MPL-1.1.txt                                *
 ***************************************************************************/
package net.sourceforge.jpcap.net;

import net.sourceforge.jpcap.util.AnsiEscapeSequences;
import net.sourceforge.jpcap.util.Timeval;
import java.nio.ByteBuffer;


/**
 * A UDP packet.
 * <p>
 * Extends an IP packet, adding a UDP header and UDP data payload.
 * Please use IPv4.UDP.Packet for new code.
 *
 * @author Patrick Charles
 * @author Jonas Lehmann
 * @author Esmond Pitt
 *
 * @version $Revision: 1.2 $
 * @lastModifiedBy $Author: esmondpitt $
 * @lastModifiedAt $Date: 2008/08/18 05:34:38 $
 */
public class UDPPacket extends IPPacket implements UDPFields, IP.UDP
{
	static final long serialVersionUID = 1080702587099492633L;
	
  /**
   * Create a new UDP packet.
   * @param lLen link-layer header length
   * @param bytes Packet data
   */
  public UDPPacket(int lLen, byte [] bytes) {
    this(lLen, bytes, null);
  }

  /**
   * Create a new UDP packet.
   * @param lLen link-layer header length
   * @param bytes Packet data
   * @param tv Timestamp
   */
  public UDPPacket(int lLen, byte [] bytes, Timeval tv) {
    this(lLen, ByteBuffer.wrap(bytes), tv);
  }

public UDPPacket(int lLen, ByteBuffer bb, Timeval tv)
{
	super(lLen, bb, tv);
}

  public int getSourcePort() {
//    return ArrayHelper.extractInteger(_bytes, _ipOffset + UDP_SP_POS,
//				      UDP_PORT_LEN);
	return _byteBuffer.getShort(_ipOffset+UDP_SP_POS) & 0xffff;
  }

  public int getDestinationPort() {
//    return ArrayHelper.extractInteger(_bytes, _ipOffset + UDP_DP_POS,
//				      UDP_PORT_LEN);
	return _byteBuffer.getShort(_ipOffset+UDP_DP_POS) & 0xffff;
  }

  /**
   * {@inheritDoc}
   *<p>
   * Fetch the total length of the UDP packet, including header and
   * data payload, in bytes.
   */
  @Override
  public int getLength() {
    // should produce the same value as header.length + data.length
//    return ArrayHelper.extractInteger(_bytes, _ipOffset + UDP_LEN_POS,
//				      UDP_LEN_LEN);
	return _byteBuffer.getShort(_ipOffset+UDP_LEN_POS) & 0xffff;
  }

  public int getUDPChecksum() {
//    return ArrayHelper.extractInteger(_bytes, _ipOffset + UDP_CSUM_POS,
//				      UDP_CSUM_LEN);
	return _byteBuffer.getShort(_ipOffset+UDP_CSUM_POS) & 0xffff;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public int getChecksum() {
    return getUDPChecksum();
  }

  public byte[] getUDPHeader() {
//    return PacketEncoding.extractHeader(_ipOffset, UDP_HEADER_LEN, _bytes);
	byte[]	bytes = new byte[UDP_HEADER_LEN];
	_byteBuffer.position(_ipOffset);
	_byteBuffer.get(bytes);
	return bytes;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public byte[] getHeader() {
    return getUDPHeader();
  }

  public byte[] getUDPData() {
      // set data length based on info in headers (note: tcpdump
      //  can return extra junk bytes which bubble up to here
//      int tmpLen = _bytes.length - _ipOffset - UDP_HEADER_LEN;
//      return PacketEncoding.extractData(_ipOffset, UDP_HEADER_LEN,
//				   _bytes, tmpLen);
	  _byteBuffer.position(_ipOffset+UDP_HEADER_LEN);
	  byte[]	bytes = new byte[_byteBuffer.remaining()];
	  _byteBuffer.get(bytes);
	  return bytes;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public byte[] getData() {
    return getUDPData();
  }

  /**
   * {@inheritDoc}
   */
  @Override
	  public boolean	isValidChecksum()
  {
	  return super.isValidChecksum() && isValidUDPChecksum();
  }
  
  public boolean	isValidUDPChecksum()
  {
	  return getUDPChecksum() == computeSenderChecksum(_ethOffset+IP_SRC_POS, getIPHeaderLength()-IP_SRC_POS+getUDPHeader().length+getUDPData().length, UDP_CSUM_POS);
  }
  
  /**
   * {@inheritDoc}
   */
  @Override
  public void	setChecksums()
  {
	  // The checksum is over all the data from the IP SRC and DST addresses to the entire UDP header and data
	  int	checksum = computeSenderChecksum(_ethOffset+IP_SRC_POS, getIPHeaderLength()-IP_SRC_POS+getUDPHeader().length+getUDPData().length, UDP_CSUM_POS);
	  _byteBuffer.putShort(_ipOffset+UDP_CSUM_POS, (short)checksum);
	  super.setChecksums();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public String toColoredString(boolean colored) {
    StringBuffer buffer = new StringBuffer();
    buffer.append('[');
    if(colored) buffer.append(getColor());
    buffer.append("UDPPacket");
    if(colored) buffer.append(AnsiEscapeSequences.RESET);
    buffer.append(": ");
    buffer.append(getSourceInetAddress().getHostAddress());
    buffer.append('.');
    buffer.append(IPPort.getName(getSourcePort()));
    buffer.append("->");
    buffer.append(getDestinationInetAddress().getHostAddress());
    buffer.append('.');
    buffer.append(IPPort.getName(getDestinationPort()));
    buffer.append(" l=" + UDP_HEADER_LEN + "," + (getLength()-UDP_HEADER_LEN));
    buffer.append(']');

    return buffer.toString();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public String getColor() {
    return AnsiEscapeSequences.LIGHT_GREEN;
  }

 private static final String _rcsid = 
    "$Id: UDPPacket.java,v 1.2 2008/08/18 05:34:38 esmondpitt Exp $";
}
